# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for closed-shell and restricted open-shell qubit mean field (QMF) ansatz. """

import unittest

import numpy as np

from tangelo.linq import Simulator
from tangelo.toolboxes.ansatz_generator.qmf import QMF
from tangelo.molecule_library import mol_H2_sto3g, mol_H4_sto3g, mol_H4_cation_sto3g

sim = Simulator()


class QMFTest(unittest.TestCase):
    """Unit tests of various functionalities of the QMF ansatz class. Examples for both closed-
    and restricted open-shell QMF are provided using H2, H4, and H4+.
    """

    @staticmethod
    def test_qmf_set_var_params():
        """ Verify behavior of set_var_params for different inputs (keyword, list, numpy array). """

        qmf_ansatz = QMF(mol_H2_sto3g)

        eight_zeros = np.zeros((8,), dtype=float)

        qmf_ansatz.set_var_params("vacuum")
        np.testing.assert_array_almost_equal(qmf_ansatz.var_params, eight_zeros, decimal=6)

        qmf_ansatz.set_var_params([0.] * 8)
        np.testing.assert_array_almost_equal(qmf_ansatz.var_params, eight_zeros, decimal=6)

        eight_pis = np.pi * np.ones((8,))

        qmf_ansatz.set_var_params("full_pi")
        np.testing.assert_array_almost_equal(qmf_ansatz.var_params, eight_pis, decimal=6)

        qmf_ansatz.set_var_params(np.array([np.pi] * 8))
        np.testing.assert_array_almost_equal(qmf_ansatz.var_params, eight_pis, decimal=6)

    def test_qmf_incorrect_number_var_params(self):
        """ Return an error if user provide incorrect number of variational parameters """

        qmf_ansatz = QMF(mol_H2_sto3g)

        self.assertRaises(ValueError, qmf_ansatz.set_var_params, np.array([1.] * 4))

    @staticmethod
    def test_qmf_set_params_upper_hf_state_h2():
        """ Verify closed-shell QMF functionalities for H2: upper case initial parameters """

        qmf_ansatz = QMF(mol_H2_sto3g)
        qmf_ansatz.set_var_params("HF_State")

        expected = np.array([np.pi] * 2 + [0.] * 6)
        np.testing.assert_allclose(expected, qmf_ansatz.var_params, rtol=1e-10)

    @staticmethod
    def test_qmf_set_params_lower_hf_state_h2():
        """ Verify closed-shell QMF functionalities for H2: lower case initial parameters """

        qmf_ansatz = QMF(mol_H2_sto3g)
        qmf_ansatz.set_var_params("hf_state")

        expected = np.array([np.pi] * 2 + [0.] * 6)
        np.testing.assert_allclose(expected, qmf_ansatz.var_params, rtol=1e-10)

    @staticmethod
    def test_qmf_set_params_hf_state_h4():
        """ Verify closed-shell QMF functionalities for H4: hf_state initial parameters """

        qmf_ansatz = QMF(mol_H4_sto3g)
        qmf_ansatz.set_var_params("hf_state")

        expected = np.array([np.pi] * 4 + [0.] * 12)
        np.testing.assert_allclose(expected, qmf_ansatz.var_params, rtol=1e-10)

    def test_qmf_closed_h2(self):
        """ Verify closed-shell QMF functionalities for H2 """

        # Build ansatz and circuit
        qmf_var_params = [np.pi] * 2 + [0.] * 6
        qmf_ansatz = QMF(mol_H2_sto3g)
        qmf_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = qmf_ansatz.qubit_ham

        # Assert energy returned is as expected for given parameters
        qmf_ansatz.update_var_params(qmf_var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, qmf_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.1166843870853400, delta=1e-6)

    def test_qmf_closed_h4(self):
        """ Verify closed-shell QMF functionalities for H4. """

        # Build ansatz and circuit
        qmf_var_params = [np.pi] * 4 + [0.] * 12
        qmf_ansatz = QMF(mol_H4_sto3g)
        qmf_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = qmf_ansatz.qubit_ham

        # Assert energy returned is as expected for given parameters
        qmf_ansatz.update_var_params(qmf_var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, qmf_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.7894832518559973, delta=1e-6)

    def test_qmf_open_h4_cation(self):
        """ Verify open-shell QMF functionalities for H4 + """

        # Build ansatz and circuit
        qmf_var_params = [np.pi] * 4 + [0.] * 12
        qmf_ansatz = QMF(mol_H4_cation_sto3g)
        qmf_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = qmf_ansatz.qubit_ham

        # Assert energy returned is as expected for given parameters
        energy = sim.get_expectation_value(qubit_hamiltonian, qmf_ansatz.circuit)
        qmf_ansatz.update_var_params(qmf_var_params)
        self.assertAlmostEqual(energy, -1.5859184313544759, delta=1e-6)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module to generate the rotations to convert the 2-body integrals into a sum of diagonal terms

Refs:
    [1] Mario Motta, Erika Ye, Jarrod R. McClean, Zhendong Li, Austin J. Minnich, Ryan Babbush, Garnet Kin-Lic Chan
    "Low rank representations for quantum simulation of electronic structure". https://arxiv.org/abs/1808.02625"""

from openfermion import givens_decomposition_square
from openfermion.circuits.low_rank import low_rank_two_body_decomposition, prepare_one_body_squared_evolution
from openfermion.chem.molecular_data import spinorb_from_spatial
import numpy as np

from tangelo.linq import Gate
from tangelo import SecondQuantizedMolecule
from tangelo.toolboxes.operators import FermionOperator
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping


def bogoliubov_transform(umat, add_phase=True):
    """Generate the gates that implement the orbital rotation defined by a given unitary.

    The given matrix is decomposed into umat=DU where U is unitary D is a diagonal matrix.
    If umat is unitary, the values of D will be exp(1j*theta) where theta is real
    and can be incorporated using add_phase=True.
    If umat is non-unitary, the values of D will be m*exp(1j*theta) where m and theta are real.
    m is ignored with only the phase (theta) incorporated in the circuit if add_phase=True.

    This implementation uses the circuit outlined in https://arxiv.org/abs/1711.05395

    Args:
        umat (array): The square unitary matrix that descibes the basis rotation
        add_phase (bool): If True, adds PHASE gates according to givens decomposition values D

    Returns:
        list of Gate: The sequence of Gates that implement the orbital rotation
    """
    gs = givens_decomposition_square(umat)
    gates = []
    if add_phase:
        for i, phase in enumerate(gs[1]):
            gates += [Gate('PHASE', i, parameter=-np.arctan2(np.imag(phase), np.real(phase)))]
    for ele in reversed(gs[0]):
        for el1 in reversed(ele):
            i, j = el1[0:2]
            gates += [Gate('CNOT', i, j), Gate('CRY', j, i, parameter=2*el1[2]), Gate('CNOT', i, j), Gate('PHASE', j, parameter=-el1[3])]

    return gates[::-1]


class OrbitalRotations():
    """The class that holds the various rotations, operators and coefficients

    Attributes:
        rotation_gates (list): The gates to apply to rotate the basis to a diagonal form for a given operator
        qubit_operators (list): The qubit operators to measure from qubit_mapping="JW" and up_then_down=False. Each is diagonal in the Z basis.
        one_body_coefficients (list): The one-body coefficients corresponding to each operator
        constants (list): The constants corresponding to each operator
        two_body_coefficients (list): The two-body coefficients corresponding to each operator
    """

    def __init__(self):
        self.rotation_gates = list()
        self.qubit_operators = list()
        self.one_body_coefficients = list()
        self.constants = list()
        self.two_body_coefficients = list()

    def add_elements(self, rot_mat, constant=0., one_body_coefs=None, two_body_coefs=None):
        """Add elements to the class for each term to diagonalize a portion of the Hamiltonian

        Args:
            rot_mat (array): The unitary rotation matrix
            constant (float): A constant value for this element
            one_body_coefs (array): The diagonal coefficients
            two_body_coefs (array): The two-body coefficients
        """
        self.rotation_gates.append(bogoliubov_transform(rot_mat))
        fe_op = FermionOperator((), coefficient=constant)
        if one_body_coefs is not None:
            for p, value in enumerate(one_body_coefs):
                fe_op += FermionOperator(((p, 1), (p, 0)), value)

        if two_body_coefs is not None:
            num_vals = np.array(two_body_coefs).shape[0]
            for p in range(num_vals):
                fe_op += FermionOperator(((p, 1), (p, 0), (p, 1), (p, 0)), two_body_coefs[p, p])
                for q in range(p+1, num_vals):
                    fe_op += FermionOperator(((p, 1), (p, 0), (q, 1), (q, 0)), two_body_coefs[p, q]/2)
                    fe_op += FermionOperator(((q, 1), (q, 0), (p, 1), (p, 0)), two_body_coefs[q, p]/2)
        self.qubit_operators.append(fermion_to_qubit_mapping(fe_op, "JW", up_then_down=False))
        self.one_body_coefficients.append(one_body_coefs)
        self.two_body_coefficients.append(two_body_coefs)
        self.constants.append(constant)

    @property
    def n_rotations(self):
        return len(self.constants)


def get_orbital_rotations(molecule: SecondQuantizedMolecule):
    """Generate the gates that rotate the orbitals such that each operator is diagonal.

    The orbital rotation gates generated are only applicable for qubit_mapping="JW" and up_then_down=False.

    Args:
        molecule (SecondQuantizedMolecule): The molecule for which the diagonal representation is requested

    Returns:
        OrbitalRotations: The class that contains the list of transformation gates and operators to measure
    """

    core_constant, one_body, two_body = molecule.get_active_space_integrals()

    eigenvalues, one_body_squares, one_body_correction, _ = low_rank_two_body_decomposition(two_body, truncation_threshold=1.e-20, spin_basis=False)

    scaled_density_density_matrices = list()
    basis_change_matrices = list()
    for j, eigenvalue in enumerate(eigenvalues):
        one_body_squares[j] = (one_body_squares[j].T.conj() + one_body_squares[j])/2
        dd_mat, basis_change_mat = prepare_one_body_squared_evolution(one_body_squares[j])
        scaled_density_density_matrices.append(np.real(eigenvalue*dd_mat))
        basis_change_matrices.append(basis_change_mat)

    # Generate basis rotation matrix and eigenvalues for one-body term
    one_body_t, _ = spinorb_from_spatial(one_body, two_body)
    one_body_coefs = one_body_t + one_body_correction
    eigs, umat = np.linalg.eigh(one_body_coefs)

    orb_rots = OrbitalRotations()
    orb_rots.add_elements(umat.T.conj(), constant=core_constant, one_body_coefs=eigs)

    # Generate elements for two-body terms
    for j in range(len(eigenvalues)):
        basis_change_mat = basis_change_matrices[j]
        two_body_coefs = scaled_density_density_matrices[j]
        orb_rots.add_elements(basis_change_mat, two_body_coefs=two_body_coefs)

    return orb_rots



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the hardware efficient ansatz class, for use in applying
VQE as first defined in "Hardware-efficient Variational Quantum Eigensolver for
Small Molecules and Quantum Magnets" https://arxiv.org/abs/1704.05018.
"""

import numpy as np

from .ansatz import Ansatz
from ._hea_circuit import construct_hea_circuit
from tangelo.toolboxes.qubit_mappings.mapping_transform import get_qubit_number
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit
from tangelo.linq import Circuit


class HEA(Ansatz):
    """This class implements the HEA ansatz. A molecule or a number of qubits +
    a number of electrons can be passed to this class.

    Args:
        molecule (SecondQuantizedMolecule) : The molecular system.
        mapping (str) : one of the supported qubit mapping identifiers. Default:
            "JW"
        up_then_down (bool): change basis ordering putting all spin up orbitals
            first, followed by all spin down. Default, False has alternating
            spin up/down ordering.
        n_layers (int): The number of HEA ansatz layers to use. One layer is
            hea_rot_type + grid of CNots. Default: 2.
        rot_type (str): "euler": RzRxRz on each qubit for each rotation layer.
            "real": Ry on each qubit for each rotation layer. Default: "euler".
        n_qubits (int) : The number of qubits in the ansatz.
            Default, None.
        n_electrons (int) : Self-explanatory.
        reference_state (str): "HF": Hartree-Fock reference state. "zero": for
            no reference state. Default: "HF".
        """

    def __init__(self, molecule=None, mapping="jw", up_then_down=False,
                n_layers=2, rot_type="euler", n_qubits=None, n_electrons=None,
                reference_state="HF"):

        if not (bool(molecule) ^ (bool(n_qubits) and (bool(n_electrons) | (reference_state == "zero")))):
            raise ValueError(f"A molecule OR qubit + electrons number must be "
                "provided when instantiating the HEA with the HF reference state. "
                "For reference_state='zero', only the number of qubits is needed.")

        if n_qubits:
            self.n_qubits = n_qubits
            self.n_electrons = n_electrons
        else:
            self.n_qubits = get_qubit_number(mapping, molecule.n_active_sos)
            self.n_electrons = molecule.n_active_electrons

        self.qubit_mapping = mapping
        self.up_then_down = up_then_down
        self.n_layers = n_layers
        self.rot_type = rot_type
        self.reference_state = reference_state

        # Each euler rotation layer has 3 variational parameters per qubit, and one non-variational entangler
        # Each real rotation layer has 1 variational parameter per qubit, and one non-variational entangler
        # There is an additional rotation layer with no entangler.
        if self.rot_type == "euler":
            self.n_var_params = self.n_qubits * 3 * (self.n_layers + 1)
        elif self.rot_type == "real":
            self.n_var_params = self.n_qubits * 1 * (self.n_layers + 1)
        else:
            raise ValueError("Supported rot_type is 'euler'' and 'real'")

        # Supported reference state initialization
        # TODO: support for others
        self.supported_reference_state = {"HF", "zero"}
        # Supported var param initialization
        self.supported_initial_var_params = {"ones", "random", "zeros"}

        # Default initial parameters for initialization
        self.var_params_default = "random"

        self.var_params = None
        self.circuit = None

    def set_var_params(self, var_params=None):
        """Set values for variational parameters, such as zeros, random numbers,
        MP2 (...), providing some keywords for users, and also supporting direct
        user input (list or numpy array). Return the parameters so that
        workflows such as VQE can retrieve these values.
        """

        if var_params is None:
            var_params = self.var_params_default

        if isinstance(var_params, str):
            if (var_params not in self.supported_initial_var_params):
                raise ValueError(f"Supported keywords for initializing variational parameters: {self.supported_initial_var_params}")
            else:
                if var_params == "ones":
                    initial_var_params = np.ones((self.n_var_params,), dtype=float)
                elif var_params == "random":
                    initial_var_params = 4 * np.pi * (np.random.random((self.n_var_params,)) - 0.5)
                elif var_params == "zeros":
                    initial_var_params = np.zeros((self.n_var_params,), dtype=float)
        else:
            initial_var_params = np.array(var_params)
            if initial_var_params.size != self.n_var_params:
                raise ValueError(f"Expected {self.n_var_params} variational parameters but "
                                 f"received {initial_var_params.size}.")
        self.var_params = initial_var_params
        return initial_var_params

    def prepare_reference_state(self):
        """Prepare a circuit generating the HF reference state."""
        if self.reference_state not in self.supported_reference_state:
            raise ValueError(f"{self.reference_state} not in supported reference state methods of:{self.supported_reference_state}")

        if self.reference_state == "HF":
            return get_reference_circuit(n_spinorbitals=self.n_qubits,
                                         n_electrons=self.n_electrons,
                                         mapping=self.qubit_mapping,
                                         up_then_down=self.up_then_down)
        elif self.reference_state == "zero":
            return get_reference_circuit(n_spinorbitals=self.n_qubits,
                                         n_electrons=0,
                                         mapping=self.qubit_mapping,
                                         up_then_down=self.up_then_down)

    def build_circuit(self, var_params=None):
        """Construct the variational circuit to be used as our ansatz."""
        self.var_params = self.set_var_params(var_params)

        reference_state_circuit = self.prepare_reference_state()

        hea_circuit = construct_hea_circuit(self.n_qubits, self.n_layers, self.rot_type)

        if reference_state_circuit.size != 0:
            self.circuit = reference_state_circuit + hea_circuit
        else:
            self.circuit = hea_circuit

        self.update_var_params(self.var_params)
        return self.circuit

    def update_var_params(self, var_params):
        """Update variational parameters (done repeatedly during VQE)."""
        self.set_var_params(var_params)
        var_params = self.var_params

        for param_index in range(self.n_var_params):
            self.circuit._variational_gates[param_index].parameter = var_params[param_index]



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides a common point for mapping operators and
 statevectors from Fermionic to Qubit encodings via any of:
 - Jordan-Wigner
 - Bravyi-Kitaev (Fenwick Tree implementation)
 - symmetry-conserving Bravyi-Kitaev (2-qubit reduction via Z2 taper)
"""


import warnings
import numpy as np
from collections.abc import Iterable
from openfermion import FermionOperator as ofFermionOperator

from tangelo.toolboxes.operators import FermionOperator, QubitOperator
from tangelo.toolboxes.qubit_mappings import jordan_wigner, bravyi_kitaev, symmetry_conserving_bravyi_kitaev, jkmn

available_mappings = {"JW", "BK", "SCBK", "JKMN"}


def get_qubit_number(mapping, n_spinorbitals):
    """Get the number of qubits for a specified number of spin orbitals, after
    passing through the qubit mapping. Symmetry-conserving Bravyi-Kitaev reduces
    the qubit number by 2.

    Args:
        mapping (string): qubit-mapping, JW, BK or SCBK.
        n_spinorbitals (int): number of spin-orbitals for molecule.

    Returns:
        int: number of qubits.
    """
    if mapping.upper() == "SCBK":
        return n_spinorbitals - 2
    else:
        return n_spinorbitals


def get_fermion_operator(operator):
    """Cast operator to FermionOperator datatype. Input is of SymbolicOperator
    type, but term words must be valid input for FermionOperator, as for example
    for InteractionOperator.

    Args:
        operator (SymbolicOperator): input operator to be cast.

    Returns:
        FermionOperator: Self-explanatory.
    """
    if not isinstance(operator, Iterable):
        raise TypeError("Input must be iterable suitable for casting to FermionOperator type.")
    fermion_operator = FermionOperator()
    for term in operator:
        try:
            fermion_operator += FermionOperator(term, operator[term])
        except Exception:
            raise TypeError("Operator terms are not formatted as valid input for FermionOperator type.")
    return fermion_operator


def fermion_to_qubit_mapping(fermion_operator, mapping, n_spinorbitals=None, n_electrons=None, up_then_down=False, spin=0):
    """Perform mapping of fermionic operator to qubit operator. This function is
    mostly a wrapper around standard openfermion code, with some important
    distinctions. We strictly enforce the specification of n_qubits for
    Bravyi-Kitaev type transformations, and n_electrons for scBK. In the absence
    of this information, these methods can return unexpected results if the
    input operator does not specifically address the highest-orbital/qubit
    index.

    Args:
        fermion_operator (FermionOperator): operator to translate to qubit
            representation.
        mapping (string): options are -- "JW" (Jordan Wigner), "BK"
            (Bravyi Kitaev), "scBK" (symmetry-conserving Bravyi Kitaev).
        n_spinorbitals (int): number of spin-orbitals for problem. Not required
            for Jordan-Wigner.
        n_electrons (int): number of occupied electron modes in problem.
            Required for symmetry conserving Bravyi-Kitaev only.
        up_then_down (bool): flag to change basis ordering, putting all spin up
            then all spin down.
        spin (int): The spin number (n_alpha - n_beta)

    Returns:
        QubitOperator: input operator, encoded in the qubit space.
    """
    # some methods may pass another operator class type. If this is the case, cast to FermionOperator where possible
    if not isinstance(fermion_operator, ofFermionOperator):
        fermion_operator = get_fermion_operator(fermion_operator)

    if mapping.upper() not in available_mappings:
        raise ValueError(f"Invalid mapping selection. Select from: {available_mappings}")

    if mapping.upper in {"BK", "SCBK", "JKMN"} and n_spinorbitals is None:
        raise ValueError(f"{mapping.upper()} requires n_spinorbitals to be set.")

    if up_then_down:
        if n_spinorbitals is None:
            raise ValueError("The number of spin-orbitals is required to execute basis re-ordering.")
        fermion_operator = make_up_then_down(fermion_operator, n_spinorbitals=n_spinorbitals)

    if mapping.upper() == "JW":
        qubit_operator = jordan_wigner(fermion_operator)
    elif mapping.upper() == "BK":
        qubit_operator = bravyi_kitaev(fermion_operator, n_qubits=n_spinorbitals)
    elif mapping.upper() == "SCBK":
        if n_electrons is None:
            raise ValueError("Symmetry-conserving Bravyi Kitaev requires specification of number of electrons.")

        qubit_operator = symmetry_conserving_bravyi_kitaev(fermion_operator=fermion_operator,
                                                           n_spinorbitals=n_spinorbitals,
                                                           n_electrons=n_electrons,
                                                           up_then_down=up_then_down,
                                                           spin=spin)
    elif mapping.upper() == "JKMN":
        qubit_operator = jkmn(fermion_operator, n_qubits=n_spinorbitals)

    converted_qubit_op = QubitOperator()
    converted_qubit_op.terms = qubit_operator.terms.copy()

    return converted_qubit_op


def make_up_then_down(fermion_operator, n_spinorbitals):
    """Re-order the orbital indexing to force all spin up, followed by all spin
        down. The default ordering alternates between spin up and down.

    Args:
        fermion_operator (FermionOperator): input operator.
        n_spinorbitals (int): number of spin-orbitals in register.

    Returns:
        FermionOperator: operator with all spin up followed by all spin down.
    """
    if not isinstance(fermion_operator, ofFermionOperator):
        raise TypeError("Invalid operator input. Must be FermionOperator.")
    if n_spinorbitals % 2 != 0:
        raise ValueError("Invalid number of spin-orbitals. Expecting even number.")
    term_modes = max([factor[0] for term in fermion_operator.terms for factor in term]) + 1
    if term_modes > n_spinorbitals:
        raise ValueError("Invalid number of modes (n_spinorbitals) for input operator. Terms in operator exceed this value.")
    new_operator = FermionOperator()

    # define the new mode register
    remapped = np.linspace(0, n_spinorbitals - 1, n_spinorbitals, dtype=int)//2
    remapped[1::2] += int(np.ceil(n_spinorbitals / 2.))

    for term, coef in fermion_operator.terms.items():

        new_term = tuple([(int(remapped[ti[0]]), ti[1]) for ti in term])
        new_operator += FermionOperator(new_term, coef)

    return new_operator



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the adaptive ansatz class."""

import math

from tangelo.linq import Circuit
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit
from tangelo.toolboxes.ansatz_generator.ansatz_utils import exp_pauliword_to_gates
from tangelo.toolboxes.ansatz_generator.ansatz import Ansatz
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping


class ADAPTAnsatz(Ansatz):
    """Adaptive ansatz used with ADAPT-VQE. This Ansatz class has methods to
    take a (or many) QubitOperator, transform it (them) into a circuit and
    append it (them). The number of variational parameters corresponds to the
    number of terms added to the Ansatz.

    Attributes:
        n_spinorbitals (int): Number of spin orbitals in a given basis.
        n_electrons (int): Number of electrons.
        operators (list of QubitOperator): List of operators to consider at the
            construction step. Can be useful for restarting computation.
        ferm_operators (list of FermionOperator): Same as operators, but in
            fermionic form. Not necessarily for running the ansatz, but it is
            convenient for analyzing results.
        mapping (string): Qubit encoding.
        up_then_down (bool): Ordering convention.
        var_params (list of float): Variational parameters.
        circuit (Circuit): Quantum circuit defined by a list of Gates.
    """

    def __init__(self, n_spinorbitals, n_electrons, ansatz_options=None):
        default_options = {"operators": list(), "ferm_operators": list(),
                           "mapping": "jw", "up_then_down": False,
                           "reference_state": "HF"}

        # Initialize with default values
        self.__dict__ = default_options
        # Overwrite default values with user-provided ones, if they correspond to a valid keyword
        if ansatz_options:
            for k, v in ansatz_options.items():
                if k in default_options:
                    setattr(self, k, v)
                else:
                    raise KeyError(f"Keyword :: {k}, not available in {self.__class__.__name__}")

        self.n_spinorbitals = n_spinorbitals
        self.n_electrons = n_electrons

        self.var_params = None
        self.circuit = None

        # The remaining of the constructor is useful to restart an ADAPT calculation.
        self._n_terms_operators = [len(pauli_op.terms) for pauli_op in self.operators]

        # Getting the sign of each pauli words. As UCCSD terms are hermitian,
        # each tau_i = excitation_i - deexcitation_i. The coefficient are
        # initialized to 1 (or -1).
        self._var_params_prefactor = list()
        for pauli_op in self.operators:
            for pauli_term in pauli_op.get_operators():
                coeff = list(pauli_term.terms.values())[0]
                self._var_params_prefactor += [math.copysign(1., coeff)]

    @property
    def n_var_params(self):
        return len(self._n_terms_operators)

    def set_var_params(self, var_params=None):
        """Set initial variational parameter values. Defaults to zeros."""

        if var_params is None:
            var_params = [0.]*self.n_var_params
        elif len(var_params) != self.n_var_params:
            raise ValueError('Invalid number of parameters.')
        self.var_params = var_params

        return var_params

    def update_var_params(self, var_params):
        """Update variational parameters (done repeatedly during VQE)."""

        for var_index in range(self.n_var_params):
            length_op = self._n_terms_operators[var_index]

            param_index = sum(self._n_terms_operators[:var_index])
            for param_subindex in range(length_op):
                prefactor = self._var_params_prefactor[param_index+param_subindex]
                self.circuit._variational_gates[param_index+param_subindex].parameter = prefactor*var_params[var_index]

    def prepare_reference_state(self):
        """Prepare a circuit generating the HF reference state."""
        if self.reference_state.upper() == "HF":
            return get_reference_circuit(n_spinorbitals=self.n_spinorbitals, n_electrons=self.n_electrons, mapping=self.mapping, up_then_down=self.up_then_down)
        else:
            return Circuit(n_qubits=self.n_spinorbitals)

    def build_circuit(self, var_params=None):
        """Construct the variational circuit to be used as our ansatz."""

        self.set_var_params(var_params)

        self.circuit = Circuit(n_qubits=self.n_spinorbitals)
        self.circuit += self.prepare_reference_state()
        adapt_circuit = list()
        for op in self.operators:
            for pauli_term in op.get_operators():
                pauli_tuple = list(pauli_term.terms.keys())[0]
                adapt_circuit += exp_pauliword_to_gates(pauli_tuple, 0.1)

        adapt_circuit = Circuit(adapt_circuit)
        if adapt_circuit.size != 0:
            self.circuit += adapt_circuit

        # Must be included because after convergence, the circuit is rebuilt.
        # If this is not present, the gradient on the previous operator chosen
        # is selected again (as the parameters are not minimized in respect to it
        # as with initialized coefficient to 0.1). If the parameters are
        # updated, it stays consistent.
        self.update_var_params(self.var_params)

        return self.circuit

    def add_operator(self, pauli_operator, ferm_operator=None):
        """Add a new operator to our circuit.

        Args:
            pauli_operator (QubitOperator): Operator to convert into a circuit
                and append it to the present circuit.
            ferm_operator (FermionicOperator): Same operator in fermionic form.
        """

        # Keeping track of the added operator.
        self._n_terms_operators += [len(pauli_operator.terms)]
        self.operators.append(pauli_operator)

        if ferm_operator is not None:
            self.ferm_operators += [ferm_operator]

        # Going through every term and convert them into a circuit. Keeping track
        # of each term sign.
        for pauli_term in pauli_operator.get_operators():
            coeff = list(pauli_term.terms.values())[0]
            self._var_params_prefactor += [math.copysign(1., coeff)]

            pauli_tuple = list(pauli_term.terms.keys())[0]
            new_operator = Circuit(exp_pauliword_to_gates(pauli_tuple, 0.1))

            self.circuit += new_operator



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class performing electronic structure calculation employing the CCSD method.
"""

from pyscf import cc, lib
from pyscf.cc.ccsd_rdm import _make_rdm1, _make_rdm2, _gamma1_intermediates, _gamma2_outcore

from tangelo.algorithms.electronic_structure_solver import ElectronicStructureSolver


class CCSDSolver(ElectronicStructureSolver):
    """Uses the CCSD method to solve the electronic structure problem, through
    pyscf.

    Args:
        molecule (SecondQuantizedMolecule): The molecule to simulate.

    Attributes:
        cc_fragment (pyscf.cc.CCSD): The coupled-cluster object.
        mean_field (pyscf.scf.RHF): The mean field of the molecule.
        frozen (list or int): Frozen molecular orbitals.
    """

    def __init__(self, molecule):
        self.cc_fragment = None

        self.mean_field = molecule.mean_field
        self.frozen = molecule.frozen_mos

    def simulate(self):
        """Perform the simulation (energy calculation) for the molecule.

        Returns:
            float: CCSD energy.
        """
        # Execute CCSD calculation
        self.cc_fragment = cc.CCSD(self.mean_field, frozen=self.frozen)
        self.cc_fragment.verbose = 0
        self.cc_fragment.conv_tol = 1e-9
        self.cc_fragment.conv_tol_normt = 1e-7

        correlation_energy, _, _ = self.cc_fragment.ccsd()
        total_energy = self.mean_field.e_tot + correlation_energy

        return total_energy

    def get_rdm(self):
        """Calculate the 1- and 2-particle reduced density matrices. The CCSD
        lambda equation will be solved for calculating the RDMs.

        Returns:
            numpy.array: One-particle RDM.
            numpy.array: Two-particle RDM.

        Raises:
            RuntimeError: If no simulation has been run.
        """

        # Check if CCSD calculation is performed
        if self.cc_fragment is None:
            raise RuntimeError("CCSDSolver: Cannot retrieve RDM. Please run the 'simulate' method first")

        # Solve the lambda equation and obtain the reduced density matrix from CC calculation
        self.cc_fragment.solve_lambda()
        t1 = self.cc_fragment.t1
        t2 = self.cc_fragment.t2
        l1 = self.cc_fragment.l1
        l2 = self.cc_fragment.l2

        d1 = _gamma1_intermediates(self.cc_fragment, t1, t2, l1, l2)
        f = lib.H5TmpFile()
        d2 = _gamma2_outcore(self.cc_fragment, t1, t2, l1, l2, f, False)

        one_rdm = _make_rdm1(self.cc_fragment, d1, with_frozen=False)
        two_rdm = _make_rdm2(self.cc_fragment, d1, d2, with_dm1=True, with_frozen=False)

        return one_rdm, two_rdm



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.algorithms.classical.ccsd_solver import CCSDSolver
from tangelo.molecule_library import mol_H2_321g, mol_Be_321g


# TODO: Can we test the get_rdm method on H2 ? How do we get our reference? Whole matrix or its properties?
class CCSDSolverTest(unittest.TestCase):

    def test_ccsd_h2(self):
        """Test CCSDSolver against result from reference implementation."""

        solver = CCSDSolver(mol_H2_321g)
        energy = solver.simulate()

        self.assertAlmostEqual(energy, -1.1478300596229851, places=6)

    def test_ccsd_be(self):
        """Test CCSDSolver against result from reference implementation."""

        solver = CCSDSolver(mol_Be_321g)
        energy = solver.simulate()

        self.assertAlmostEqual(energy, -14.531416589890926, places=6)

    def test_ccsd_be_frozen_core(self):
        """ Test CCSDSolver against result from reference implementation, with
        no mean-field provided as input. Frozen core is considered.
        """

        mol_Be_321g_freeze1 = mol_Be_321g.freeze_mos(1, inplace=False)

        solver = CCSDSolver(mol_Be_321g_freeze1)
        energy = solver.simulate()

        self.assertAlmostEqual(energy, -14.530687987160581, places=6)

    def test_ccsd_be_as_two_levels(self):
        """ Test CCSDSolver against result from reference implementation, with
        no mean-field provided as input. This atom is reduced to an HOMO-LUMO
        problem.
        """

        mol_Be_321g_freeze_list = mol_Be_321g.freeze_mos([0, 3, 4, 5, 6, 7, 8], inplace=False)

        solver = CCSDSolver(mol_Be_321g_freeze_list)
        energy = solver.simulate()

        self.assertAlmostEqual(energy, -14.498104489160106, places=6)

    def test_ccsd_get_rdm_without_simulate(self):
        """Test that the runtime error is raised when user calls get RDM without
        first running a simulation.
        """

        solver = CCSDSolver(mol_H2_321g)
        self.assertRaises(RuntimeError, solver.get_rdm)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fragment class, used in construction of ONIOM layers -- contains details of
both the constituent geometry (i.e. which atoms from system are in fragment,
which bonds are broken and how to fix them) as well as the solver(s) to use.
"""

import warnings

import numpy as np
from scipy.spatial.transform import Rotation as R

from tangelo import SecondQuantizedMolecule
from tangelo.algorithms import CCSDSolver, FCISolver, VQESolver, MINDO3Solver, ADAPTSolver, QITESolver
from tangelo.problem_decomposition.oniom._helpers.capping_groups import elements, chemical_groups


class Fragment:

    def __init__(self, solver_low, options_low=None, solver_high=None, options_high=None, selected_atoms=None, charge=0, spin=0, broken_links=None):
        """Fragment class for the ONIOM solver. Each fragment can have broken
        links. In this case, they are capped with a chosen atom. Each fragment
        can also have up to two solvers (low and high accuracy).

        Args:
            solver_low (str): Specification of low accuracy solver for fragment.
            options_low (dict): Specification of low accuracy solver options.
            solver_high (str): Specification of higher accuracy solver for
                fragment.
            options_high (dict): Specification of higher accuracy solver options.
            selected_atoms (list of int or int): Which atoms from molecule are
                in fragment. int counts from start of xyz.
            spin (int): Spin associated witht this fragment.
            charge (int): Charge associated witht this fragment.
            broken_links (list of Link): Bonds broken when forming fragment.

        Attributes:
            mol_low (SecondQuantizedMolecule): Molecule of this fragment, to be
                solved with low accuracy.
            mol_high (SecondQuantizedMolecule): Molecule of this fragment, to be
                solved with high accuracy.
            e_fragment (float): Energy of this fragment as defined by ONIOM.
                None if simulate has not yet been called.
        """

        default_solver_options = {"basis": "sto-3g"}

        # Check to see if a fragment has no solver_high when only a portion of a molecule
        # is selected. If this is allowed, the energy of a fragment is added to the
        # system (all system with solver_low + fragment with solver_low), adding
        # more atoms in the system than there are.
        if selected_atoms is not None and solver_high is None:
            raise RuntimeError("If atoms are selected (selected_atoms different than None), a solver_high must be provided.")

        self.selected_atoms = selected_atoms

        # Solver with low accuracy.
        self.solver_low = solver_low.upper()
        self.options_low = options_low if options_low is not None else default_solver_options

        # Solver with higher accuracy.
        self.solver_high = solver_high.upper() if solver_high is not None else solver_high
        self.options_high = options_high if options_high is not None else default_solver_options

        self.supported_classical_solvers = {"HF": None, "CCSD": CCSDSolver, "FCI": FCISolver, "MINDO3": MINDO3Solver}
        self.supported_quantum_solvers = {"VQE": VQESolver, "ADAPT": ADAPTSolver, "QITE": QITESolver}

        # Check if the solvers are implemented in ONIOM.
        builtin_solvers = self.supported_classical_solvers.keys() | self.supported_quantum_solvers.keys()
        if self.solver_low not in builtin_solvers:
            raise NotImplementedError(f"This {self.solver_low} solver has not been implemented yet in {self.__class__.__name__}")
        elif self.solver_high and self.solver_high not in builtin_solvers:
            raise NotImplementedError(f"This {self.solver_high} solver has not been implemented yet in {self.__class__.__name__}")

        # For this fragment (not the whole molecule).
        self.spin = spin
        self.charge = charge

        self.broken_links = broken_links
        self.mol_low = None
        self.mol_high = None
        self.e_fragment = None

    def build(self):
        """Get the solver objects for this layer. Also defined molecule objects."""

        # Low accuracy solver.
        # We begin by defining the molecule.
        if self.mol_low is None:
            self.mol_low = self.get_mol(self.options_low["basis"], self.options_low.get("frozen_orbitals", None))
            # Basis is only relevant when computing the mean-field. After this,
            # it is discarded (not needed for electronic solver because they
            # retrieved it from the molecule object).
            self.options_low = {i: self.options_low[i] for i in self.options_low if i not in ["basis", "frozen_orbitals"]}

        self.solver_low = self.get_solver(self.mol_low, self.solver_low, self.options_low)

        # Higher accuracy solver.
        if self.solver_high is not None:

            # Molecule is reconstructed (in case a different basis is used).
            if self.mol_high is None:
                self.mol_high = self.get_mol(self.options_high["basis"], self.options_high.get("frozen_orbitals", None))
                # Same process done as in low accuracy process.
                self.options_high = {i: self.options_high[i] for i in self.options_high if i not in ["basis", "frozen_orbitals"]}

            self.solver_high = self.get_solver(self.mol_high, self.solver_high, self.options_high)

    def simulate(self):
        """Get the energy for this fragment.

        Returns:
            float: Energy for the fragment.
        """

        # Low accuracy solver.
        e_low = Fragment.get_energy(self.mol_low, self.solver_low)

        # Higher accuracy solver.
        e_high = 0.
        if self.solver_high is not None:
            e_high = Fragment.get_energy(self.mol_high, self.solver_high)

            # Contribution from low accuracy is substracted, as defined by ONIOM.
            e_low *= -1

        self.e_fragment = e_high + e_low
        return self.e_fragment

    def get_mol(self, basis, frozen=None):
        """Get the molecule object for this fragment (with a specified basis).

        Returns:
            SecondQuantizedMolecule: Molecule object.
        """

        return SecondQuantizedMolecule(self.geometry, self.charge, self.spin, basis, frozen_orbitals=frozen)

    @staticmethod
    def get_energy(molecule, solver):
        """Get the energy for a specific solver.

        Args:
            molecule (SecondQuantizedMolecule): Molecule for this fragment (with
                repaired links).
            solver (solver object or string): Which solver to use.

        Returns:
            float: Energy of the fragment.
        """

        # In case of RHF solver (inside SecondQuantizedMolecule object).
        if isinstance(solver, str):
            energy = molecule.mf_energy
        # The remaining case is for VQESolver, CCSDSolver, FCISolver and
        # MINDO3Solver.
        else:
            energy = solver.simulate()

        return energy

    def get_solver(self, molecule, solver_string, options_solver):
        """Get the solver object (or string for RHF) for this layer.

        Args:
            molecule (SecondQuantizedMolecule): Molecule for this fragment (with
                repaired links).
            solver_string (str): Which solver to use.
            options_solver (dict): Options for the solver.

        Returns:
            ElectronicStructureSolver: Solver object or string.
        """

        if solver_string == "HF":
            return "HF"
        elif solver_string in self.supported_classical_solvers:
            return self.supported_classical_solvers[solver_string](molecule, **options_solver)
        elif solver_string in self.supported_quantum_solvers:
            molecule_options = {"molecule": molecule}
            solver = self.supported_quantum_solvers[solver_string]({**molecule_options, **options_solver})
            solver.build()
            return solver

    def get_resources(self):
        """ Estimate the quantum esources required for this fragment Only
        supports VQESolver solvers.
        """

        # If one of the solver is VQE, quantum resources are returned. If both
        # are VQE, the solver_high overrides the solver_low resources.
        resources = {}

        quantum_solvers = tuple(self.supported_quantum_solvers.values())

        if isinstance(self.solver_low, quantum_solvers):
            resources = self.solver_low.get_resources()

        if isinstance(self.solver_high, quantum_solvers):
            resources = self.solver_high.get_resources()

        return resources


class Link:

    def __init__(self, staying, leaving, factor=1.0, species="H"):
        """Bonds broken during the layer-construction process in ONIOM must be
        mended. This class represents a broken-bond link, and has associated
        methods to generate a new bond, appending the intended species.

        Args:
            staying (int): Atom id retained.
            leaving (int): Atom id lost.
            factor (float) optional: Rescale length of bond, from that in the
                original molecule.
            species (str) optional: Atomic species or a chemical group
                identifier for new link. Can be a list (first element = "X" to
                detect the orientation) for a custom chemical group.
        """

        self.staying = staying
        self.leaving = leaving
        self.factor = factor

        if isinstance(species, str) and species in elements:
            self.species = [(species, (0., 0., 0.))]
        elif isinstance(species, str) and species in chemical_groups:
            self.species = chemical_groups[species]
        elif isinstance(species, (list, tuple)) and species[0][0].upper() == "X":
            self.species = species
        else:
            raise ValueError(f"{species} is not supported. It must be a string identifier or a list of atoms (with a ghost atom ('X') as the first element).")

    def relink(self, geometry):
        """Create atom at location of mended-bond link.

        Args:
            geometry (list of positions): Atomic positions in format
                [[str,tuple(float,float,float)],...].

        Returns:
            list: List of atomic species and position (x, y, z) of replacement
                atom / chemical group.
        """

        elements = [a[0] for a in self.species if a[0].upper() != "X"]
        chem_group_xyz = np.array([[a[1][0], a[1][1], a[1][2]] for a in self.species if a[0].upper() != "X"])

        staying = np.array(geometry[self.staying][1])
        leaving = np.array(geometry[self.leaving][1])

        # Rotation (if not a single atom).
        if len(elements) > 1:
            axis_old = leaving - staying
            axis_new = chem_group_xyz[0] - np.array(self.species[0][1])

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                rot, _ = R.align_vectors([axis_old], [axis_new])
            chem_group_xyz = rot.apply(chem_group_xyz)

        # Move the atom / group to the right position in space.
        replacement = self.factor*(leaving-staying) + staying
        translation = replacement - chem_group_xyz[0]
        chem_group_xyz += translation

        return [(element, (xyz[0], xyz[1], xyz[2])) for element, xyz in zip(elements, chem_group_xyz)]



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from scipy.special import binom

from tangelo.toolboxes.ansatz_generator._general_unitary_cc import *


class UCCGSDTest(unittest.TestCase):

    def test_spin_order(self):
        """Test that spin-ordering is implemented correctly, for both
        openfermion and qiskit orderings.
        """
        n_orbitals = 6
        p, q, r, s = 0, 1, 2, 3
        qiskit = np.array([0, 1, 2, 3]), 6 + np.array([0, 1, 2, 3])
        fermion = np.array([0, 2, 4, 6]), np.array([1, 3, 5, 7])
        up_q, dn_q = get_spin_ordered(n_orbitals, p, q, r, s, up_down=True)
        up_f, dn_f = get_spin_ordered(n_orbitals, p, q, r, s, up_down=False)

        self.assertEqual(np.linalg.norm(up_q - qiskit[0]), 0.0, msg="Spin Up Qiskit-Ordering Fails")
        self.assertEqual(np.linalg.norm(dn_q - qiskit[1]), 0.0, msg="Spin Down Qiskit-Ordering Fails")
        self.assertEqual(np.linalg.norm(up_f - fermion[0]), 0.0, msg="Spin Up openfermion-Ordering Fails")
        self.assertEqual(np.linalg.norm(dn_f - fermion[1]), 0.0, msg="Spin Down openfermion-Ordering Fails")

    def count_doubles_groups(self, n_orbs, up_down=False):
        """General test for number of doubles groups found by generator."""
        selection = np.linspace(0, n_orbs - 1, n_orbs, dtype=int)
        groups = np.zeros(5)
        for pp, qq, rr, ss in itertools.product(selection, repeat=4):

            if (pp < qq and pp < rr and pp < ss) and (rr < ss) and (qq != rr and qq != ss):
                _ = get_group_1_2(n_orbs, pp, qq, rr, ss, up_down=up_down)
                groups[0] += 1
            elif qq == rr and pp < ss and pp != qq and ss != qq:
                _ = get_group_1_2(n_orbs, pp, qq, rr, ss, up_down=up_down)
                groups[1] += 1
            elif (pp == qq and qq != rr and rr != ss and ss != pp) and rr < ss:
                _ = get_group_3_4(n_orbs, pp, qq, rr, ss, up_down=up_down)
                groups[2] += 1
            elif pp == qq and qq == rr and rr != ss:
                _ = get_group_3_4(n_orbs, pp, qq, rr, ss, up_down=up_down)
                groups[3] += 1
            elif pp == qq and qq != rr and rr == ss and pp < ss:
                _ = get_group_5(n_orbs, pp, qq, rr, ss, up_down=up_down)
                groups[4] += 1

        self.assertEqual(groups[0], binom(n_orbs, 2) * binom(n_orbs - 2, 2) // 2,
                         msg="{:d} orbs: Invalid Group 1 Number".format(n_orbs))
        self.assertEqual(groups[1], n_orbs * binom(n_orbs - 1, 2),
                         msg="{:d} orbs: Invalid Group 2 Number".format(n_orbs))
        self.assertEqual(groups[2], n_orbs * binom(n_orbs - 1, 2),
                         msg="{:d} orbs: Invalid Group 3 Number".format(n_orbs))
        self.assertEqual(groups[3], 2 * binom(n_orbs, 2), msg="{:d} orbs: Invalid Group 4 Number".format(n_orbs))
        self.assertEqual(groups[4], binom(n_orbs, 2), msg="{:d} orbs: Invalid Group 5 Number".format(n_orbs))
        self.assertEqual(np.sum(groups), get_doubles_number(n_orbs),
                         msg="{:d} orbs: Invalid Total Number".format(n_orbs))

    def test_count_doubles(self):
        """Test for checking number of doubles excitations generated."""
        self.count_doubles_groups(5, up_down=False)
        self.count_doubles_groups(5, up_down=True)
        self.count_doubles_groups(8, up_down=False)

    def test_count_singles(self):
        """Test for checking number of singles excitations generated."""
        self.assertEqual(len(get_singles(5)), get_singles_number(5), msg="5 orbs: Invalid Number of singles")
        self.assertEqual(len(get_singles(10)), get_singles_number(10), msg="5 orbs: Invalid Number of singles")

    def test_coeff_pass(self):
        """Test that errors are handled correctly when an invalid number of
        coefficients are passed. Also tests that good input returns good output.
        """
        n_qubits = 4

        # Get expected number of single- and double-coefficients
        n_single_coeffs = get_singles_number(n_qubits // 2)
        n_double_coeffs = get_doubles_number(n_qubits // 2)

        # Generate good coefficients
        good_single_coeffs = np.random.random(n_single_coeffs)
        good_double_coeffs = np.random.random(n_double_coeffs)
        # Generate bad coefficients
        bad_single_coeffs = np.random.random(n_single_coeffs + 10)
        bad_double_coeffs = np.random.random(n_double_coeffs + 10)

        # Combine singles and doubles
        good_coeffs = np.concatenate((good_single_coeffs, good_double_coeffs))

        # Check that an invalid input raises a ValueError in each possible case
        with self.assertRaises(ValueError):
            get_coeffs(n_qubits, bad_single_coeffs, good_double_coeffs)
        with self.assertRaises(ValueError):
            get_coeffs(n_qubits, bad_single_coeffs, good_double_coeffs)
        with self.assertRaises(ValueError):
            get_coeffs(n_qubits, bad_single_coeffs, bad_double_coeffs)

        # Check that a valid input gives the expected result
        good_result = get_coeffs(n_qubits, good_single_coeffs, good_double_coeffs)
        self.assertTrue(np.allclose(good_coeffs, good_result))


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements a collection of functions related to the QMF
ansatz:
    1. Analytical evaluation of an expectation value of a QubitOperator
       using a QMF wave function;
    2. Initialization of the QMF variational parameter set {Omega} from a
       Hartree-Fock reference state;
    3. Purification {Omega} when building and screening the DIS of QCC generators;
    4. Construction of a QMF state circuit using {Omega};
    5. Addition of terms for N, S^2, and Sz that penalize a mean-field Hamiltonian
       in order to obtain solutions corresponding to specific electron number
       and spin symmetries.
For more information, see references below.

Refs:
    1. I. G. Ryabinkin and S. N. Genin.
        https://arxiv.org/abs/1812.09812 2018.
    2. S. N. Genin, I. G. Ryabinkin, and A. F. Izmaylov.
          https://arxiv.org/abs/1901.04715 2019.
    3. I. G. Ryabinkin, S. N. Genin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2019, 15, 1, 249–255.
"""

import numpy as np

from tangelo.linq import Circuit, Gate
from tangelo.toolboxes.operators.operators import FermionOperator
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_vector, get_mapped_vector
from tangelo.toolboxes.ansatz_generator.penalty_terms import combined_penalty, number_operator_penalty,\
                                                             spin2_operator_penalty, spin_operator_penalty


def get_op_expval(qubit_op, qmf_var_params):
    """Driver function for analytical evaluation of a QubitOperator expectation value with a
    QMF wave function.

    Args:
        qubit_op (QubitOperator): A qubit operator to compute the expectation value of.
        qmf_var_params (numpy array of float): QMF variational parameter set.

    Returns:
        complex: expectation value of all qubit operator terms.
    """

    n_qubits = qmf_var_params.size // 2
    qubit_op_gen = ((qop_items[0], (qop_items[1], qmf_var_params, n_qubits))
                     for qop_items in qubit_op.terms.items())
    return sum([calc_op_expval(qop_gen[0], *qop_gen[1]) for qop_gen in qubit_op_gen])


def calc_op_expval(qop_term, *qop_qmf_data):
    """Analytically evaluate a qubit operator expectation value with a QMF wave function.
    The expectation values of Pauli operators X, Y, and Z are (Ref. 2)
    <QMF| X | QMF> = cos(phi) * sin(theta), <QMF| Y | QMF> = sin(phi) * sin(theta),
    <QMF| Z | QMF> = cos(theta)

    Args:
        qop_term (tuple of tuple): The Pauli operators and indices of a QubitOperator term.
        qop_qmf_data (tuple): The coefficient of a QubitOperator term, QMF variational parameter
            set (numpy array of float), and number of qubits (int).

    Returns:
        complex: expectation value of a qubit operator term.
    """

    coef, qmf_var_params, n_qubits = qop_qmf_data
    for idx, pauli in qop_term:
        theta, phi = qmf_var_params[idx], qmf_var_params[idx + n_qubits]
        if pauli == "X":
            coef *= np.cos(phi) * np.sin(theta)
        elif pauli == "Y":
            coef *= np.sin(phi) * np.sin(theta)
        elif pauli == "Z":
            coef *= np.cos(theta)
    return coef


def init_qmf_from_hf(n_spinorbitals, n_electrons, mapping, up_then_down=False, spin=None):
    """Function to initialize the QMF variational parameter set from a Hartree-Fock state
    occupation vector. The theta Bloch angles are set to 0. or np.pi if the molecular orbital is
    unoccupied or occupied, respectively. The phi Bloch angles are set to 0.

    Args:
        n_spinorbitals (int): Number of spin-orbitals in the molecular system.
        n_electrons (int): Number of electrons in the molecular system.
        mapping (str) : One of the supported qubit mapping identifiers.
        up_then_down (bool): Change basis ordering putting all spin-up orbitals first,
            followed by all spin-down.
        spin (int): 2*S = n_alpha - n_beta.

    Returns:
        numpy array of float: QMF variational parameter set.
    """

    # Get thetas from HF vec and arrange Bloch angles so all thetas are first then phis
    thetas = get_vector(n_spinorbitals, n_electrons, mapping, up_then_down, spin)
    return np.concatenate((np.pi * thetas, np.zeros((len(thetas),), dtype=float)))


def init_qmf_from_vector(vector, mapping, up_then_down=False):
    """Function to initialize the QMF variational parameter set from a Hartree-Fock state
    occupation vector. The theta Bloch angles are set to 0. or np.pi if the molecular orbital is
    unoccupied or occupied, respectively. The phi Bloch angles are set to 0.

    Args:
        vector (array): Occupation vector of orbitals using alternating up and down electrons (i.e. up_then_down=False)
        mapping (str): One of the supported qubit mapping identifiers.
        up_then_down (bool): Change basis ordering putting all spin-up orbitals first,
            followed by all spin-down when applying the qubit mapping.

    Returns:
        numpy array of float: QMF variational parameter set.
    """

    # Get thetas from HF vec and arrange Bloch angles so all thetas are first then phis
    thetas = np.array(get_mapped_vector(vector, mapping, up_then_down))
    var_params = np.zeros(2*len(thetas))
    var_params[:len(thetas)] = thetas
    return var_params


def purify_qmf_state(qmf_var_params, n_spinorbitals, n_electrons, mapping, up_then_down=False, spin=None):
    """The efficient construction and screening of the DIS requires a z-collinear QMF state.
    If the QMF state specified by qmf_var_params is not z-collinear, this function adjusts the
    parameters to the nearest z-collinear computational basis state.

    Args:
        qmf_var_params (numpy array of float): QMF variational parameter set.
        n_spinorbitals (int): Number of spin-orbitals in the molecular system.
        n_electrons (int): Number of electrons in the molecular system.
        mapping (str) : One of the supported qubit mapping identifiers.
        up_then_down (bool): Change basis ordering putting all spin-up orbitals first,
            followed by all spin-down.
        spin (int): 2*S = n_alpha - n_beta.

    Returns:
        numpy array of float: purified QMF parameter set that corresponds to the
            nearest z-collinear state to the current QMF state.
    """

    # Adjust the theta Bloch angles
    pure_var_params, n_qubits = np.copy(qmf_var_params), qmf_var_params.size // 2
    for i, theta in enumerate(qmf_var_params[:n_qubits]):
        c_0, c_1 = np.cos(0.5 * theta), np.sin(0.5 * theta)
        if abs(c_0) > abs(c_1):
            pure_var_params[i] = 0.
        elif abs(c_0) < abs(c_1):
            pure_var_params[i] = np.pi
        else:
            vector = get_vector(n_spinorbitals, n_electrons, mapping, up_then_down, spin)
            pure_var_params[i] = np.pi * vector[i]
    return pure_var_params


def get_qmf_circuit(qmf_var_params, variational=True):
    """Build a QMF state preparation circuit using the current state of the QMF variational
    parameter set {Omega}. The first n_qubit elements in {Omega} are parameters for RX gates,
    and the second n_qubit elements in {Omega} are parameters for RZ gates.

    Args:
        qmf_var_params (numpy array of float): QMF variational parameter set.
        variational (bool): Flag to treat {Omega} variationally or keep them fixed.

    Returns:
        Circuit: instance of tangelo.linq Circuit class.
    """

    n_qubits, gates = qmf_var_params.size // 2, []
    for idx, param in enumerate(qmf_var_params):
        gate_id = "RX" if idx < n_qubits else "RZ"
        gates.append(Gate(gate_id, target=idx % n_qubits, parameter=param, is_variational=variational))
    return Circuit(gates)


def penalize_mf_ham(mf_pen_terms, n_orbitals):
    """Generate a FermionOperator that is used to penalize a mean-field Hamiltonian for at
    least one of the N, S^2, or Sz operators.

    Args:
        mf_pen_terms (dict): Parameters for mean-field Hamiltonian penalization.
            The keys are "N", "S^2", or "Sz" (str) and the values are a tuple of the penalty
            term coefficient, mu (float), and the target value of a penalty operator (int).
            Example - "key": (mu, target). Key, value pairs are case sensitive and mu > 0.
        n_orbitals (int): Number of orbitals in the fermion basis.

    Returns:
        FermionOperator: sum of all mean-field Hamiltonian penalty terms.
    """

    mf_penalty = FermionOperator.zero()
    penalty_terms = list(key for key in mf_pen_terms.keys())
    if len(penalty_terms) == 3:
        mf_penalty += combined_penalty(n_orbitals, mf_pen_terms, False)
    else:
        for penalty_term in penalty_terms:
            coef, target = mf_pen_terms[penalty_term]
            if coef <= 0.:
                raise ValueError("The penalty term coefficient must be positive.")
            if penalty_term == "N":
                mf_penalty += number_operator_penalty(n_orbitals, target, coef, False)
            elif penalty_term == "S^2":
                mf_penalty += spin2_operator_penalty(n_orbitals, target, coef, False)
            elif penalty_term == "Sz":
                mf_penalty += spin_operator_penalty(n_orbitals, target, coef, False)
    return mf_penalty



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the variational quantum eigensolver (VQE) algorithm to solve
electronic structure calculations.
"""

import warnings
import itertools

from enum import Enum
import numpy as np
from openfermion.ops.operators.qubit_operator import QubitOperator

from tangelo.helpers.utils import HiddenPrints
from tangelo.linq import Simulator, Circuit
from tangelo.linq.helpers.circuits.measurement_basis import measurement_basis_gates
from tangelo.toolboxes.operators import count_qubits, FermionOperator, qubitop_to_qubitham
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_mapped_vector, vector_to_circuit
from tangelo.toolboxes.ansatz_generator.ansatz import Ansatz
from tangelo.toolboxes.ansatz_generator import UCCSD, RUCC, HEA, UpCCGSD, QMF, QCC, VSQS, UCCGD,  ILC,\
                                               VariationalCircuitAnsatz
from tangelo.toolboxes.ansatz_generator._qubit_mf import init_qmf_from_vector
from tangelo.toolboxes.ansatz_generator.penalty_terms import combined_penalty
from tangelo.toolboxes.post_processing.bootstrapping import get_resampled_frequencies
from tangelo.toolboxes.ansatz_generator.fermionic_operators import number_operator, spinz_operator, spin2_operator
from tangelo.toolboxes.optimizers.rotosolve import rotosolve


class BuiltInAnsatze(Enum):
    """Enumeration of the ansatz circuits supported by VQE."""
    UCCSD = 0
    UCC1 = 1
    UCC3 = 2
    HEA = 3
    UpCCGSD = 4
    QMF = 5
    QCC = 6
    VSQS = 7
    UCCGD = 8
    ILC = 9


class VQESolver:
    r"""Solve the electronic structure problem for a molecular system by using
    the variational quantum eigensolver (VQE) algorithm.

    This algorithm evaluates the energy of a molecular system by performing
    classical optimization over a parametrized ansatz circuit.

    Users must first set the desired options of the VQESolver object through the
    __init__ method, and call the "build" method to build the underlying objects
    (mean-field, hardware backend, ansatz...). They are then able to call any of
    the energy_estimation, simulate, or get_rdm methods. In particular, simulate
    runs the VQE algorithm, returning the optimal energy found by the classical
    optimizer.

    Attributes:
        molecule (SecondQuantizedMolecule) : the molecular system.
        qubit_mapping (str) : one of the supported qubit mapping identifiers.
        ansatz (Ansatze) : one of the supported ansatze.
        optimizer (function handle): a function defining the classical optimizer
            and its behavior.
        initial_var_params (str or array-like) : initial value for the classical
            optimizer.
        backend_options (dict) : parameters to build the tangelo.linq Simulator
            class.
        penalty_terms (dict): parameters for penalty terms to append to target
            qubit Hamiltonian (see penalty_terms for more details).
        deflation_circuits (list[Circuit]): Deflation circuits to add an
            orthogonalization penalty with.
        deflation_coeff (float): The coefficient of the deflation.
        ansatz_options (dict): parameters for the given ansatz (see given ansatz
            file for details).
        up_then_down (bool): change basis ordering putting all spin up orbitals
            first, followed by all spin down. Default, False has alternating
                spin up/down ordering.
        qubit_hamiltonian (QubitOperator-like): Self-explanatory.
        verbose (bool): Flag for VQE verbosity.
        ref_state (array or Circuit): The reference configuration to use. Replaces HF state
            QMF, QCC, ILC require ref_state to be an array. UCC1, UCC3, VSQS can not use a
            different ref_state than HF by construction.
    """

    def __init__(self, opt_dict):

        default_backend_options = {"target": None, "n_shots": None, "noise_model": None}
        default_options = {"molecule": None,
                           "qubit_mapping": "jw", "ansatz": BuiltInAnsatze.UCCSD,
                           "optimizer": self._default_optimizer,
                           "initial_var_params": None,
                           "backend_options": default_backend_options,
                           "penalty_terms": None,
                           "deflation_circuits": list(),
                           "deflation_coeff": 1,
                           "ansatz_options": dict(),
                           "up_then_down": False,
                           "qubit_hamiltonian": None,
                           "verbose": False,
                           "ref_state": None}

        # Initialize with default values
        self.__dict__ = default_options
        # Overwrite default values with user-provided ones, if they correspond to a valid keyword
        for k, v in opt_dict.items():
            if k in default_options:
                setattr(self, k, v)
            else:
                raise KeyError(f"Keyword :: {k}, not available in VQESolver")

        # Raise error/warnings if input is not as expected. Only a single input
        # must be provided to avoid conflicts.
        if not (bool(self.molecule) ^ bool(self.qubit_hamiltonian)):
            raise ValueError(f"A molecule OR qubit Hamiltonian object must be provided when instantiating {self.__class__.__name__}.")

        # The QCC & ILC ansatze require up_then_down=True when mapping="jw"
        if isinstance(self.ansatz, BuiltInAnsatze):
            if self.ansatz in (BuiltInAnsatze.QCC, BuiltInAnsatze.ILC) and self.qubit_mapping.lower() == "jw" and not self.up_then_down:
                warnings.warn("Spin-orbital ordering shifted to all spin-up first then down to ensure efficient generator screening "
                              "for the Jordan-Wigner mapping with QCC-based ansatze.", RuntimeWarning)
                self.up_then_down = True
            # QCC and QMF and ILC require a reference state that can be represented by a single layer of RZ-RX gates on each qubit.
            # This decomposition can not be determined from a general Circuit reference state.
            if isinstance(self.ref_state, Circuit):
                if self.ansatz in [BuiltInAnsatze.QCC, BuiltInAnsatze.ILC, BuiltInAnsatze.QMF]:
                    raise ValueError("Circuit reference state is not supported for QCC or QMF")
            elif self.ref_state is not None:
                if self.ansatz in [BuiltInAnsatze.QCC, BuiltInAnsatze.ILC]:
                    self.ansatz_options["qmf_var_params"] = init_qmf_from_vector(self.ref_state, self.qubit_mapping, self.up_then_down)
                    self.ref_state = None
                elif self.ansatz == BuiltInAnsatze.QMF:
                    self.ansatz_options["init_qmf"] = {"init_params": "vector", "vector": self.ref_state}
                    self.ref_state = None
                # UCC1, UCC3, QMF and VSQS require the initial state to be Hartree-Fock.
                # UCC1 and UCC3 use a special structure
                # VSQS is only defined for a Hartree-Fock reference at this time
                elif self.ansatz in [BuiltInAnsatze.UCC1, BuiltInAnsatze.UCC3, BuiltInAnsatze.VSQS]:
                    raise ValueError("UCC1, UCC3, and VSQS do not support reference states other than Hartree-Fock at this time in Tangelo")

        if self.ref_state is not None:
            if isinstance(self.ref_state, Circuit):
                self.reference_circuit = self.ref_state
            else:
                self.reference_circuit = vector_to_circuit(get_mapped_vector(self.ref_state, self.qubit_mapping, self.up_then_down))
        else:
            self.reference_circuit = Circuit()

        self.default_backend_options = default_backend_options
        self.optimal_energy = None
        self.optimal_var_params = None
        self.builtin_ansatze = set(BuiltInAnsatze)

    def build(self):
        """Build the underlying objects required to run the VQE algorithm
        afterwards.
        """

        if isinstance(self.ansatz, Circuit):
            self.ansatz = VariationalCircuitAnsatz(self.ansatz)

        # Check compatibility of optimizer with Ansatz class
        elif self.optimizer == rotosolve:
            if self.ansatz not in [BuiltInAnsatze.UCC1, BuiltInAnsatze.UCC3, BuiltInAnsatze.HEA]:
                raise ValueError(f"{self.ansatz} not compatible with rotosolve optimizer.")

        # Building VQE with a molecule as input.
        if self.molecule:

            # Compute qubit hamiltonian for the input molecular system
            qubit_op = fermion_to_qubit_mapping(fermion_operator=self.molecule.fermionic_hamiltonian,
                                                mapping=self.qubit_mapping,
                                                n_spinorbitals=self.molecule.n_active_sos,
                                                n_electrons=self.molecule.n_active_electrons,
                                                up_then_down=self.up_then_down,
                                                spin=self.molecule.spin)

            self.qubit_hamiltonian = qubitop_to_qubitham(qubit_op, self.qubit_mapping, self.up_then_down)

            if self.penalty_terms:
                pen_ferm = combined_penalty(self.molecule.n_active_mos, self.penalty_terms)
                pen_qubit = fermion_to_qubit_mapping(fermion_operator=pen_ferm,
                                                     mapping=self.qubit_mapping,
                                                     n_spinorbitals=self.molecule.n_active_sos,
                                                     n_electrons=self.molecule.n_active_electrons,
                                                     up_then_down=self.up_then_down,
                                                     spin=self.molecule.spin)
                pen_qubit = qubitop_to_qubitham(pen_qubit, self.qubit_hamiltonian.mapping, self.qubit_hamiltonian.up_then_down)
                self.qubit_hamiltonian += pen_qubit

            # Verification of system compatibility with UCC1 or UCC3 circuits.
            if self.ansatz in [BuiltInAnsatze.UCC1, BuiltInAnsatze.UCC3]:
                # Mapping should be JW because those ansatz are chemically inspired.
                if self.qubit_mapping.upper() != "JW":
                    raise ValueError("Qubit mapping must be JW for {} ansatz.".format(self.ansatz))
                # They are encoded with this convention.
                if not self.up_then_down:
                    raise ValueError("Parameter up_then_down must be set to True for {} ansatz.".format(self.ansatz))
                # Only HOMO-LUMO systems are relevant.
                if count_qubits(self.qubit_hamiltonian) != 4:
                    raise ValueError("The system must be reduced to a HOMO-LUMO problem for {} ansatz.".format(self.ansatz))

            # Build / set ansatz circuit. Use user-provided circuit or built-in ansatz depending on user input.
            if isinstance(self.ansatz, BuiltInAnsatze):
                if self.ansatz == BuiltInAnsatze.UCCSD:
                    self.ansatz = UCCSD(self.molecule, self.qubit_mapping, self.up_then_down)
                    self.ansatz.default_reference_state = "HF" if self.ref_state is None else "zero"
                elif self.ansatz == BuiltInAnsatze.UCC1:
                    self.ansatz = RUCC(1)
                elif self.ansatz == BuiltInAnsatze.UCC3:
                    self.ansatz = RUCC(3)
                elif self.ansatz == BuiltInAnsatze.HEA:
                    if self.ref_state is not None:
                        self.ansatz_options["reference_state"] = "zero"
                    self.ansatz = HEA(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                elif self.ansatz == BuiltInAnsatze.UpCCGSD:
                    self.ansatz = UpCCGSD(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                    self.ansatz.default_reference_state = "HF" if self.ref_state is None else "zero"
                elif self.ansatz == BuiltInAnsatze.QMF:
                    self.ansatz = QMF(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                elif self.ansatz == BuiltInAnsatze.QCC:
                    self.ansatz = QCC(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                elif self.ansatz == BuiltInAnsatze.VSQS:
                    self.ansatz = VSQS(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                elif self.ansatz == BuiltInAnsatze.UCCGD:
                    self.ansatz = UCCGD(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                    self.ansatz.default_reference_state = "HF" if self.ref_state is None else "zero"
                elif self.ansatz == BuiltInAnsatze.ILC:
                    self.ansatz = ILC(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                else:
                    raise ValueError(f"Unsupported ansatz. Built-in ansatze:\n\t{self.builtin_ansatze}")
            elif not isinstance(self.ansatz, Ansatz):
                raise TypeError(f"Invalid ansatz dataype. Expecting instance of Ansatz class, or one of built-in options:\n\t{self.builtin_ansatze}")

        # Building with a qubit Hamiltonian.
        elif self.ansatz in [BuiltInAnsatze.HEA, BuiltInAnsatze.VSQS]:
            if self.ansatz == BuiltInAnsatze.HEA:
                self.ansatz = HEA(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
            elif self.ansatz == BuiltInAnsatze.VSQS:
                self.ansatz = VSQS(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
        elif not isinstance(self.ansatz, Ansatz):
            raise TypeError(f"Invalid ansatz dataype. Expecting a custom Ansatz (Ansatz class).")

        # Set ansatz initial parameters (default or use input), build corresponding ansatz circuit
        self.initial_var_params = self.ansatz.set_var_params(self.initial_var_params)
        self.ansatz.build_circuit()

        if len(self.ansatz.circuit._variational_gates) == 0:
            warnings.warn("No variational gate found in the circuit.", RuntimeWarning)

        # Quantum circuit simulation backend options
        t = self.backend_options.get("target", self.default_backend_options["target"])
        ns = self.backend_options.get("n_shots", self.default_backend_options["n_shots"])
        nm = self.backend_options.get("noise_model", self.default_backend_options["noise_model"])
        self.backend = Simulator(target=t, n_shots=ns, noise_model=nm)

    def simulate(self):
        """Run the VQE algorithm, using the ansatz, classical optimizer, initial
        parameters and hardware backend built in the build method.
        """
        if not (self.ansatz and self.backend):
            raise RuntimeError("No ansatz circuit or hardware backend built. Have you called VQESolver.build ?")

        if len(self.ansatz.circuit._variational_gates) == 0:
            raise RuntimeError("No variational gate found in the circuit.")

        optimal_energy, optimal_var_params = self.optimizer(self.energy_estimation, self.initial_var_params)

        self.optimal_var_params = optimal_var_params
        self.optimal_energy = optimal_energy
        self.ansatz.build_circuit(self.optimal_var_params)
        self.optimal_circuit = self.reference_circuit+self.ansatz.circuit if self.ref_state is not None else self.ansatz.circuit
        return self.optimal_energy

    def get_resources(self):
        """Estimate the resources required by VQE, with the current ansatz. This
        assumes "build" has been run, as it requires the ansatz circuit and the
        qubit Hamiltonian. Return information that pertains to the user, for the
        purpose of running an experiment on a classical simulator or a quantum
        device.
        """

        resources = dict()
        resources["qubit_hamiltonian_terms"] = len(self.qubit_hamiltonian.terms)
        resources["circuit_width"] = self.ansatz.circuit.width
        resources["circuit_gates"] = self.ansatz.circuit.size
        # For now, only CNOTs supported.
        resources["circuit_2qubit_gates"] = self.ansatz.circuit.counts.get("CNOT", 0)
        resources["circuit_var_gates"] = len(self.ansatz.circuit._variational_gates)
        resources["vqe_variational_parameters"] = len(self.initial_var_params)
        return resources

    def energy_estimation(self, var_params):
        """Estimate energy using the given ansatz, qubit hamiltonian and compute
        backend. Keeps track of optimal energy and variational parameters along
        the way.

        Args:
             var_params (numpy.array or str): variational parameters to use for
                VQE energy evaluation.

        Returns:
             float: energy computed by VQE using the ansatz and input
                variational parameters.
        """

        # Update variational parameters, compute energy using the hardware backend
        self.ansatz.update_var_params(var_params)
        circuit = self.ansatz.circuit if self.ref_state is None else self.reference_circuit + self.ansatz.circuit
        energy = self.backend.get_expectation_value(self.qubit_hamiltonian, circuit)

        # Additional computation for deflation (optional)
        for circ in self.deflation_circuits:
            f_dict, _ = self.backend.simulate(circ + circuit.inverse())
            energy += self.deflation_coeff * f_dict.get("0"*self.ansatz.circuit.width, 0)

        if self.verbose:
            print(f"\tEnergy = {energy:.7f} ")

        return energy

    def operator_expectation(self, operator, var_params=None, n_active_mos=None, n_active_electrons=None, n_active_sos=None, spin=None, ref_state=Circuit()):
        """Obtains the operator expectation value of a given operator.

           Args:
                operator (str or QubitOperator): The operator to find the
                    expectation value of str availability:
                    - N : Particle number
                    - Sz: Spin in z-direction
                    - S^2: Spin quantum number s(s+1)
                var_params (str or numpy.array): variational parameters to use
                    for VQE expectation value evaluation.
                n_active_mos (int): The number of active_mos (int). Only
                    required when using a str input and VQESolver is initiated
                    with a QubitHamiltonian.
                n_active_electrons (int): The number of active electrons. Only
                    required when operator is of type FermionOperator and
                    mapping used is scbk and vqe_solver was initiated using a
                    QubitHamiltonian.
                n_active_sos (int): Number of active spin orbitals. Only
                    required when operator is of type FermionOperator and
                    mapping used is scbk and vqe_solver was initiated using a
                    QubitHamiltonian.
                spin (int): Spin (n_alpha - n_beta)
                ref_state (Circuit): A reference state preparation circuit

           Returns:
                float: operator expectation value computed by VQE using the
                    ansatz and input variational parameters.
        """
        if var_params is None:
            var_params = self.ansatz.var_params

        # Save our current target hamiltonian
        tmp_hamiltonian = self.qubit_hamiltonian

        if isinstance(operator, str):
            if n_active_mos is None:
                if self.molecule:
                    n_active_mos = self.molecule.n_active_mos
                else:
                    raise KeyError("Must supply n_active_mos when a QubitHamiltonian has initialized VQESolver"
                                   " and requesting the expectation of 'N', 'Sz', or 'S^2'")
            if operator == "N":
                exp_op = number_operator(n_active_mos, up_then_down=False)
            elif operator == "Sz":
                exp_op = spinz_operator(n_active_mos, up_then_down=False)
            elif operator == "S^2":
                exp_op = spin2_operator(n_active_mos, up_then_down=False)
            else:
                raise ValueError('Only expectation values of N, Sz and S^2')
        elif isinstance(operator, FermionOperator):
            exp_op = operator
        elif isinstance(operator, QubitOperator):
            self.qubit_hamiltonian = operator
        else:
            raise TypeError("operator must be a of string, FermionOperator or QubitOperator type.")

        if isinstance(operator, (str, FermionOperator)):
            if (n_active_electrons is None or n_active_sos is None or spin is None) and self.qubit_hamiltonian.mapping == "scbk":
                if self.molecule:
                    n_active_electrons = self.molecule.n_active_electrons
                    n_active_sos = self.molecule.n_active_sos
                    spin = self.molecule.spin
                else:
                    raise KeyError("Must supply n_active_electrons, n_active_sos, and spin with a FermionOperator and scbk mapping.")

            self.qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=exp_op,
                                                              mapping=self.qubit_hamiltonian.mapping,
                                                              n_spinorbitals=n_active_sos,
                                                              n_electrons=n_active_electrons,
                                                              up_then_down=self.qubit_hamiltonian.up_then_down,
                                                              spin=spin)

        self.ansatz.update_var_params(var_params)
        expectation = self.backend.get_expectation_value(self.qubit_hamiltonian, ref_state+self.ansatz.circuit)

        # Restore the current target hamiltonian
        self.qubit_hamiltonian = tmp_hamiltonian

        return expectation

    def get_rdm(self, var_params, resample=False, sum_spin=True, ref_state=Circuit()):
        """Compute the 1- and 2- RDM matrices using the VQE energy evaluation.
        This method allows to combine the DMET problem decomposition technique
        with the VQE as an electronic structure solver. The RDMs are computed by
        using each fermionic Hamiltonian term, transforming them and computing
        the elements one-by-one. Note that the Hamiltonian coefficients will not
        be multiplied as in the energy evaluation. The first element of the
        Hamiltonian is the nuclear repulsion energy term, not the Hamiltonian
        term.

        Args:
            var_params (numpy.array or list): variational parameters to use for
                rdm calculation
            resample (bool): Whether to resample saved frequencies. get_rdm with
                savefrequencies=True must be called or a dictionary for each
                qubit terms' frequencies must be set to self.rdm_freq_dict
            sum_spin (bool): If True, the spin-summed 1-RDM and 2-RDM will be
                returned. If False, the full 1-RDM and 2-RDM will be returned.
            ref_state (Circuit): A reference state preparation circuit.

        Returns:
            (numpy.array, numpy.array): One & two-particle spin summed RDMs if
                sumspin=True or the full One & two-Particle RDMs if
                sumspin=False.
        """

        self.ansatz.update_var_params(var_params)

        # Initialize the RDM arrays
        n_mol_orbitals = self.molecule.n_active_mos
        n_spin_orbitals = self.molecule.n_active_sos
        rdm1_spin = np.zeros((n_spin_orbitals,) * 2, dtype=complex)
        rdm2_spin = np.zeros((n_spin_orbitals,) * 4, dtype=complex)

        # If resampling is requested, check that a previous savefrequencies run has been called
        if resample:
            if hasattr(self, "rdm_freq_dict"):
                qb_freq_dict = self.rdm_freq_dict
                resampled_expect_dict = dict()
            else:
                raise AttributeError("Need to run RDM calculation with savefrequencies=True")
        else:
            qb_freq_dict = dict()
            qb_expect_dict = dict()

        # Loop over each element of Hamiltonian (non-zero value)
        for key in self.molecule.fermionic_hamiltonian.terms:
            # Ignore constant / empty term
            if not key:
                continue

            # Assign indices depending on one- or two-body term
            length = len(key)
            if (length == 2):
                iele, jele = (int(ele[0]) for ele in tuple(key[0:2]))
            elif (length == 4):
                iele, jele, kele, lele = (int(ele[0]) for ele in tuple(key[0:4]))

            # Create the Hamiltonian with the correct key (Set coefficient to one)
            hamiltonian_temp = FermionOperator(key)

            # Obtain qubit Hamiltonian
            qubit_hamiltonian2 = fermion_to_qubit_mapping(fermion_operator=hamiltonian_temp,
                                                          mapping=self.qubit_mapping,
                                                          n_spinorbitals=self.molecule.n_active_sos,
                                                          n_electrons=self.molecule.n_active_electrons,
                                                          up_then_down=self.up_then_down,
                                                          spin=self.molecule.spin)
            qubit_hamiltonian2.compress()

            # Run through each qubit term separately, use previously calculated result for the qubit term or
            # calculate and save results for that qubit term
            opt_energy2 = 0.
            for qb_term, qb_coef in qubit_hamiltonian2.terms.items():
                if qb_term:
                    if qb_term not in qb_freq_dict:
                        if resample:
                            warnings.warn(f"Warning: rerunning circuit for missing qubit term {qb_term}")
                        basis_circuit = Circuit(measurement_basis_gates(qb_term))
                        full_circuit = ref_state + self.ansatz.circuit + basis_circuit
                        qb_freq_dict[qb_term], _ = self.backend.simulate(full_circuit)
                    if resample:
                        if qb_term not in resampled_expect_dict:
                            resampled_freq_dict = get_resampled_frequencies(qb_freq_dict[qb_term], self.backend.n_shots)
                            resampled_expect_dict[qb_term] = self.backend.get_expectation_value_from_frequencies_oneterm(qb_term, resampled_freq_dict)
                        expectation = resampled_expect_dict[qb_term]
                    else:
                        if qb_term not in qb_expect_dict:
                            qb_expect_dict[qb_term] = self.backend.get_expectation_value_from_frequencies_oneterm(qb_term, qb_freq_dict[qb_term])
                        expectation = qb_expect_dict[qb_term]
                    opt_energy2 += qb_coef * expectation
                else:
                    opt_energy2 += qb_coef

            # Put the values in np arrays (differentiate 1- and 2-RDM)
            if length == 2:
                rdm1_spin[iele, jele] += opt_energy2
            elif length == 4:
                rdm2_spin[iele, lele, jele, kele] += opt_energy2

        # save rdm frequency dictionary
        self.rdm_freq_dict = qb_freq_dict

        if sum_spin:
            rdm1_np = np.zeros((n_mol_orbitals,) * 2, dtype=np.complex128)
            rdm2_np = np.zeros((n_mol_orbitals,) * 4, dtype=np.complex128)

            # Construct spin-summed 1-RDM
            for i, j in itertools.product(range(n_spin_orbitals), repeat=2):
                rdm1_np[i//2, j//2] += rdm1_spin[i, j]

            # Construct spin-summed 2-RDM
            for i, j, k, l in itertools.product(range(n_spin_orbitals), repeat=4):
                rdm2_np[i//2, j//2, k//2, l//2] += rdm2_spin[i, j, k, l]

            return rdm1_np, rdm2_np

        return rdm1_spin, rdm2_spin

    def _default_optimizer(self, func, var_params):
        """Function used as a default optimizer for VQE when user does not
        provide one. Can be used as an example for users who wish to provide
        their custom optimizer.

        Should set the attributes "optimal_var_params" and "optimal_energy" to
        ensure the outcome of VQE is captured at the end of classical
        optimization, and can be accessed in a standard way.

        Args:
            func (function handle): The function that performs energy
                estimation. This function takes var_params as input and returns
                a float.
            var_params (list): The variational parameters (float64).

        Returns:
            float: The optimal energy found by the optimizer.
            list of floats: Optimal parameters
        """

        from scipy.optimize import minimize

        with HiddenPrints():
            result = minimize(func, var_params, method="SLSQP",
                              options={"disp": True, "maxiter": 2000, "eps": 1e-5, "ftol": 1e-5})

        if self.verbose:
            print(f"VQESolver optimization results:")
            print(f"\tOptimal VQE energy: {result.fun}")
            print(f"\tOptimal VQE variational parameters: {result.x}")
            print(f"\tNumber of Iterations : {result.nit}")
            print(f"\tNumber of Function Evaluations : {result.nfev}")
            print(f"\tNumber of Gradient Evaluations : {result.njev}")

        return result.fun, result.x



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides functions allowing users to estimate the number of
measurements that are needed to approximate the expectation values of some qubit
operators up to a given accuracy.

The functions range from simple general heuristics to more complex approaches
that may leverage our knowledge of the quantum state.
"""

import math


def get_measurement_estimate(qb_op, digits=3, method="uniform"):
    """Given a qubit operator and a level of accuracy, computes the number of
    measurements required by each term of the qubit operator to reach the
    accuracy provided by the user when computing expectation values, returns it
    as a dictionary mapping measurement basis to number of measurements.

    "uniform" method makes no assumption about the underlying probability
    distribution resulting from the quantum state preparation circuit. The rule
    of thumb is "Multiply number of samples by 100 for each digit of accuracy".

    Args:
        qb_op: qubit operator.
        digits (integer): number of digits of accuracy desired on expectation
            value.

    Returns:
        dict: Dictionary mapping terms / measurement bases to their number of
            measurements.
    """

    available_methods = {'uniform'}
    if method not in available_methods:
        raise NotImplementedError(f"Only available methods are {available_methods}")

    scaling_factor = 100**(digits+1)

    measurements = dict()
    for term, coef in qb_op.terms.items():
        coef = max(abs(coef.real), abs(coef.imag))
        if not term or coef < 10**(-digits):
            measurements[term] = 0
        else:
            n_samples = math.floor(scaling_factor * 100**(math.log10(coef)))
            # Assign to dictionary, handle edge case
            measurements[term] = 100 if n_samples == 1 else n_samples

    return measurements



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions helping with quantum circuit format conversion between abstract
format and qiskit format.

In order to produce an equivalent circuit for the target backend, it is
necessary to account for:
- how the gate names differ between the source backend to the target backend.
- how the order and conventions for some of the inputs to the gate operations
    may also differ.
"""


def get_qiskit_gates():
    """Map gate name of the abstract format to the equivalent add_gate method of
    Qiskit's QuantumCircuit class API and supported gates:
    https://qiskit.org/documentation/stubs/qiskit.circuit.QuantumCircuit.html
    """

    import qiskit

    GATE_QISKIT = dict()
    GATE_QISKIT["H"] = qiskit.QuantumCircuit.h
    GATE_QISKIT["X"] = qiskit.QuantumCircuit.x
    GATE_QISKIT["Y"] = qiskit.QuantumCircuit.y
    GATE_QISKIT["Z"] = qiskit.QuantumCircuit.z
    GATE_QISKIT["CH"] = qiskit.QuantumCircuit.ch
    GATE_QISKIT["CX"] = qiskit.QuantumCircuit.cx
    GATE_QISKIT["CY"] = qiskit.QuantumCircuit.cy
    GATE_QISKIT["CZ"] = qiskit.QuantumCircuit.cz
    GATE_QISKIT["S"] = qiskit.QuantumCircuit.s
    GATE_QISKIT["T"] = qiskit.QuantumCircuit.t
    GATE_QISKIT["RX"] = qiskit.QuantumCircuit.rx
    GATE_QISKIT["RY"] = qiskit.QuantumCircuit.ry
    GATE_QISKIT["RZ"] = qiskit.QuantumCircuit.rz
    GATE_QISKIT["CRX"] = qiskit.QuantumCircuit.crx
    GATE_QISKIT["CRY"] = qiskit.QuantumCircuit.cry
    GATE_QISKIT["CRZ"] = qiskit.QuantumCircuit.crz
    GATE_QISKIT["CNOT"] = qiskit.QuantumCircuit.cx
    GATE_QISKIT["SWAP"] = qiskit.QuantumCircuit.swap
    GATE_QISKIT["XX"] = qiskit.QuantumCircuit.rxx
    GATE_QISKIT["CSWAP"] = qiskit.QuantumCircuit.cswap
    GATE_QISKIT["PHASE"] = qiskit.QuantumCircuit.p
    GATE_QISKIT["CPHASE"] = qiskit.QuantumCircuit.cp
    GATE_QISKIT["MEASURE"] = qiskit.QuantumCircuit.measure
    return GATE_QISKIT


def translate_qiskit(source_circuit):
    """Take in an abstract circuit, return an equivalent qiskit QuantumCircuit
    instance

    Args:
        source_circuit: quantum circuit in the abstract format.

    Returns:
        qiskit.QuantumCircuit: the corresponding qiskit quantum circuit.
    """

    import qiskit

    GATE_QISKIT = get_qiskit_gates()
    target_circuit = qiskit.QuantumCircuit(source_circuit.width, source_circuit.width)

    # Maps the gate information properly. Different for each backend (order, values)
    for gate in source_circuit._gates:
        if gate.control is not None:
            if len(gate.control) > 1:
                raise ValueError('Multi-controlled gates not supported with qiskit. Gate {gate.name} with controls {gate.control} is not allowed')
        if gate.name in {"H", "X", "Y", "Z", "S", "T"}:
            (GATE_QISKIT[gate.name])(target_circuit, gate.target[0])
        elif gate.name in {"RX", "RY", "RZ", "PHASE"}:
            (GATE_QISKIT[gate.name])(target_circuit, gate.parameter, gate.target[0])
        elif gate.name in {"CRX", "CRY", "CRZ", "CPHASE"}:
            (GATE_QISKIT[gate.name])(target_circuit, gate.parameter, gate.control[0], gate.target[0])
        elif gate.name in {"CNOT", "CH", "CX", "CY", "CZ"}:
            (GATE_QISKIT[gate.name])(target_circuit, gate.control[0], gate.target[0])
        elif gate.name in {"SWAP"}:
            (GATE_QISKIT[gate.name])(target_circuit, gate.target[0], gate.target[1])
        elif gate.name in {"CSWAP"}:
            (GATE_QISKIT[gate.name])(target_circuit, gate.control[0], gate.target[0], gate.target[1])
        elif gate.name in {"XX"}:
            (GATE_QISKIT[gate.name])(target_circuit, gate.parameter, gate.target[0], gate.target[1])
        elif gate.name in {"MEASURE"}:
            (GATE_QISKIT[gate.name])(target_circuit, gate.target[0], gate.target[0])
        else:
            raise ValueError(f"Gate '{gate.name}' not supported on backend qiskit")
    return target_circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for performing Bravyi-Kitaev Transformation, as prescribed via the
Fenwick Tree mapping. This implementation accommodates mapping of qubit
registers where the number of qubits is not a power of two.

NB: this is a minimal implementation, just wrapping around current openfermion
code (v 1.0.1). This wrapper enables future refactoring to utilize our own
implementation, as needed.
"""

from openfermion.utils import count_qubits
from openfermion.transforms import bravyi_kitaev as openfermion_bravyi_kitaev


def bravyi_kitaev(fermion_operator, n_qubits):
    """Execute transformation of FermionOperator to QubitOperator using the
    Bravyi-Kitaev transformation. Important note: there are several
    implementations of "Bravyi Kitaev" transformation, in both the literature,
    and historical versions of openfermion. This function executes the
    transformaton defined in arXiv:quant-ph/0003137. Different versions are not
    necessarily the same, and result in undesirable performance. This method is
    a simple wrapper around openfermion's bravyi_kitaev, but we are forcing the
    user to pass n_qubits to avoid unexpected behaviour.

    Args:
        fermion_operator (FermionOperator): input fermionic operator to be
            transformed.
        n_qubits (int): number of qubits associated with the operator.

    Returns:
        QubitOperator: output bravyi-kitaev encoded qubit operator.
    """
    if not (type(n_qubits) is int):
        raise TypeError("Number of qubits (n_qubits) must be integer type.")
    if n_qubits < count_qubits(fermion_operator):
        raise ValueError("Invalid (too few) number of qubits (n_qubits) for input operator.")

    qubit_operator = openfermion_bravyi_kitaev(fermion_operator, n_qubits=n_qubits)

    return qubit_operator



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

import numpy as np
from openfermion.linalg import qubit_operator_sparse
from openfermion.utils import load_operator

from tangelo.toolboxes.operators import QubitOperator
from tangelo.toolboxes.operators.taper_qubits import QubitTapering

# For openfermion.load_operator function.
pwd_this_test = os.path.dirname(os.path.abspath(__file__))


class QubitTaperingTest(unittest.TestCase):

    def test_unsupported_mapping(self):
        """Test unsupported mapping for qubit for tapering."""

        with self.assertRaises(NotImplementedError):
            QubitTapering(QubitOperator(), 4, 2, 0, "scBK", True)

    def test_z2taper_h2_jw_occupied_first(self):
        """Test Z2 tapering of H2 JW up_then_down=False."""

        qu_op = load_operator("H2_JW_occfirst.data", data_directory=pwd_this_test+"/data", plain_text=True)
        e = np.min(np.linalg.eigvalsh(qubit_operator_sparse(qu_op).todense()))

        tapering = QubitTapering(qu_op, 4, 2, 0, "JW", False)
        tapered_qu_op = tapering.z2_tapered_op.qubitoperator
        e_taper = np.min(np.linalg.eigvalsh(qubit_operator_sparse(tapered_qu_op).todense()))

        self.assertAlmostEqual(e, e_taper, places=5)

    def test_z2taper_h2_jw_spinup_first(self):
        """Test Z2 tapering of H2 JW up_then_down=True."""

        qu_op = load_operator("H2_JW_spinupfirst.data", data_directory=pwd_this_test+"/data", plain_text=True)
        e = np.min(np.linalg.eigvalsh(qubit_operator_sparse(qu_op).todense()))

        tapering = QubitTapering(qu_op, 4, 2, 0, "JW", True)
        tapered_qu_op = tapering.z2_tapered_op.qubitoperator
        e_taper = np.min(np.linalg.eigvalsh(qubit_operator_sparse(tapered_qu_op).todense()))

        self.assertAlmostEqual(e, e_taper, places=5)

    def test_z2taper_h2_bk_occupied_first(self):
        """Test Z2 tapering of H2 BK up_then_down=False."""

        qu_op = load_operator("H2_BK_occfirst.data", data_directory=pwd_this_test+"/data", plain_text=True)
        e = np.min(np.linalg.eigvalsh(qubit_operator_sparse(qu_op).todense()))

        tapering = QubitTapering(qu_op, 4, 2, 0, "BK", False)
        tapered_qu_op = tapering.z2_tapered_op.qubitoperator
        e_taper = np.min(np.linalg.eigvalsh(qubit_operator_sparse(tapered_qu_op).todense()))

        self.assertAlmostEqual(e, e_taper, places=5)

    def test_z2taper_h2_bk_spinup_first(self):
        """Test Z2 tapering of H2 BK up_then_down=True."""

        qu_op = load_operator("H2_BK_spinupfirst.data", data_directory=pwd_this_test+"/data", plain_text=True)
        e = np.min(np.linalg.eigvalsh(qubit_operator_sparse(qu_op).todense()))

        tapering = QubitTapering(qu_op, 4, 2, 0, "BK", True)
        tapered_qu_op = tapering.z2_tapered_op.qubitoperator
        e_taper = np.min(np.linalg.eigvalsh(qubit_operator_sparse(tapered_qu_op).todense()))

        self.assertAlmostEqual(e, e_taper, places=5)

    def test_z2taper_h2_jkmn_occupied_first(self):
        """Test Z2 tapering of H2 JKMN up_then_down=False."""

        qu_op = load_operator("H2_JKMN_occfirst.data", data_directory=pwd_this_test+"/data", plain_text=True)
        e = np.min(np.linalg.eigvalsh(qubit_operator_sparse(qu_op).todense()))

        tapering = QubitTapering(qu_op, 4, 2, 0, "JKMN", False)
        tapered_qu_op = tapering.z2_tapered_op.qubitoperator
        e_taper = np.min(np.linalg.eigvalsh(qubit_operator_sparse(tapered_qu_op).todense()))

        self.assertAlmostEqual(e, e_taper, places=5)

    def test_z2taper_h2_jkmn_spinup_first(self):
        """Test Z2 tapering of H2 JKMN up_then_down=True."""

        qu_op = load_operator("H2_JKMN_spinupfirst.data", data_directory=pwd_this_test+"/data", plain_text=True)
        e = np.min(np.linalg.eigvalsh(qubit_operator_sparse(qu_op).todense()))

        tapering = QubitTapering(qu_op, 4, 2, 0, "JKMN", True)
        tapered_qu_op = tapering.z2_tapered_op.qubitoperator
        e_taper = np.min(np.linalg.eigvalsh(qubit_operator_sparse(tapered_qu_op).todense()))

        self.assertAlmostEqual(e, e_taper, places=5)

    def test_z2taper_h2_triplet(self):
        """Test Z2 tapering of H2 triplet JW up_then_down=False."""

        qu_op = load_operator("H2_JW_occfirst.data", data_directory=pwd_this_test+"/data", plain_text=True)

        # Triplet state is the second lowest energy after the groud state.
        e = np.sort(np.linalg.eigvalsh(qubit_operator_sparse(qu_op).todense()))[1]

        tapering = QubitTapering(qu_op, 4, 2, 2, "JW", False)
        tapered_qu_op = tapering.z2_tapered_op.qubitoperator
        e_taper = np.min(np.linalg.eigvalsh(qubit_operator_sparse(tapered_qu_op).todense()))
        self.assertAlmostEqual(e, e_taper, places=5)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.toolboxes.operators import QubitHamiltonian, QubitOperator, count_qubits, qubitop_to_qubitham


class OperatorsUtilitiesTest(unittest.TestCase):

    def test_count_qubits(self):
        """Test count_qubits function."""

        n_qubits_3 = count_qubits(QubitHamiltonian("X0 Y1 Z2", mapping="JW", up_then_down=True))
        self.assertEqual(n_qubits_3, 3)

        n_qubits_5 = count_qubits(QubitHamiltonian("X0 Y1 Z4", mapping="JW", up_then_down=True))
        self.assertEqual(n_qubits_5, 5)

    def test_qubitop_to_qubitham(self):
        """Test qubitop_to_qubitham function."""

        qubit_ham = qubitop_to_qubitham(0.5*QubitOperator("X0 Y1 Z2"), "BK", False)

        reference_attributes = {"terms": {((0, 'X'), (1, 'Y'), (2, 'Z')): 0.5},
                                "mapping": "BK", "up_then_down": False}

        self.assertDictEqual(qubit_ham.__dict__, reference_attributes)


class QubitHamiltonianTest(unittest.TestCase):

    def test_instantiate_QubitHamiltonian(self):
        """Test initialization of QubitHamiltonian class."""

        qubit_ham = 0.5 * QubitHamiltonian("X0 Y1 Z2", mapping="BK", up_then_down=False)
        reference_attributes = {"terms": {((0, 'X'), (1, 'Y'), (2, 'Z')): 0.5},
                                "mapping": "BK", "up_then_down": False}

        self.assertDictEqual(qubit_ham.__dict__, reference_attributes)

    def test_instantiate_QubitHamiltonian(self):
        """Test error raising when 2 incompatible QubitHamiltonian are summed up
        together.
        """

        qubit_ham = QubitHamiltonian("X0 Y1 Z2", mapping="JW", up_then_down=True)

        with self.assertRaises(RuntimeError):
            qubit_ham + QubitHamiltonian("Z0 X1 Y2", mapping="BK", up_then_down=True)

        with self.assertRaises(RuntimeError):
            qubit_ham + QubitHamiltonian("Z0 X1 Y2", mapping="JW",  up_then_down=False)

    def test_to_qubit_operator(self):
        """Test exportation of QubitHamiltonian to QubitOperator."""

        qubit_ham = 1. * QubitHamiltonian("X0 Y1 Z2", mapping="JW",  up_then_down=True)

        self.assertEqual(qubit_ham.to_qubitoperator(),
            QubitOperator(term="X0 Y1 Z2", coefficient=1.))

    def test_get_operators(self):
        """Test get_operators methods, defined in QubitOperator class."""

        terms = [QubitHamiltonian("X0 Y1 Z2", mapping="JW", up_then_down=True),
                 QubitHamiltonian("Z0 X1 Y2", mapping="JW", up_then_down=True),
                 QubitHamiltonian("Y0 Z1 X2", mapping="JW", up_then_down=True)]

        H = terms[0] + terms[1] + terms[2]

        for ref_term, term in zip(terms, H.get_operators()):
            self.assertEqual(ref_term, term)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" This module defines an ansatz class to wrap up a custom tangelo.linq
    circuit.
"""

import numpy as np

from tangelo.toolboxes.ansatz_generator.ansatz import Ansatz


class VariationalCircuitAnsatz(Ansatz):
    """This class takes an arbitrary circuit and convert it to an Ansatz. This
    enables users to provide a custom pre-built circuit.

    Args:
        Circuit: Circuit with variational gates.
    """

    def __init__(self, abstract_circuit):

        self.circuit = abstract_circuit

        self.n_var_params = len(self.circuit._variational_gates)
        self.var_params = None

        # Supported var param initialization
        self.supported_initial_var_params = {"ones", "random", "zeros"}

        # Default initial parameters for initialization
        self.var_params_default = [gate.parameter for gate in self.circuit._variational_gates]

    def set_var_params(self, var_params=None):
        """Set initial variational parameter values. Defaults to random."""

        if var_params is None:
            var_params = self.var_params_default

        if isinstance(var_params, str):
            var_params = var_params.lower()

            if (var_params not in self.supported_initial_var_params):
                raise ValueError(f"Supported keywords for initializing variational parameters: {self.supported_initial_var_params}")
            else:
                if var_params == "ones":
                    var_params = np.ones((self.n_var_params,), dtype=float)
                elif var_params == "random":
                    var_params = 4 * np.pi * (np.random.random((self.n_var_params,)) - 0.5)
                elif var_params == "zeros":
                    var_params = np.zeros((self.n_var_params,), dtype=float)
        else:
            var_params = np.array(var_params)
            if var_params.size != self.n_var_params:
                raise ValueError(f"Expected {self.n_var_params} variational parameters but "
                                 f"received {var_params.size}.")
        self.var_params = var_params
        return var_params

    def update_var_params(self, var_params):
        """Update variational parameters (done repeatedly during VQE)."""

        self.set_var_params(var_params)
        var_params = self.var_params

        for param_index in range(self.n_var_params):
            self.circuit._variational_gates[param_index].parameter = var_params[param_index]

    def prepare_reference_state(self):
        """Method not needed as it is expected to be in the circuit provided."""
        pass

    def build_circuit(self, var_params=None):
        """Update parameters of the pre-built circuit."""

        self.set_var_params(var_params)
        self.update_var_params(self.var_params)

        return self.circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Our own n-layered Integrated molecular Orbital and Molecular mechanics
(ONIOM) solver. User specifies either the number (if beginning from start of
list), or the indices, of atoms which are to be identified as the model
system(s), from the larger molecular system.

Main model class for running oniom-calculations. This is analogous to the
scf.RHF, etc. methods, requiring however a bit more information. User supplies
an atomic-geometry, and specifies the system, as well as necessary models,
of increasing sophistication.

Reference:
    - The ONIOM Method and Its Applications. Lung Wa Chung, W. M. C. Sameera,
    Romain Ramozzi, Alister J. Page, Miho Hatanaka, Galina P. Petrova,
    Travis V. Harris, Xin Li, Zhuofeng Ke, Fengyi Liu, Hai-Bei Li, Lina Ding
    and Keiji Morokuma
    Chemical Reviews 2015 115 (12), 5678-5796. DOI: 10.1021/cr5004419.
"""

from tangelo.problem_decomposition.problem_decomposition import ProblemDecomposition
from tangelo.toolboxes.molecular_computation.molecule import atom_string_to_list


class ONIOMProblemDecomposition(ProblemDecomposition):

    def __init__(self, opt_dict):
        """Main class for the ONIOM hybrid solver. It defines layers with
        different electronic solvers.

        Attributes:
            geometry (strin or list): XYZ atomic coords (in "str float float..."
                or [[str, (float, float, float)], ...] format).
            fragments (list of Fragment): Specification of different
                system-subgroups and their solvers.
            verbose (bolean): Verbose flag.
        """

        default_options = {"geometry": None,
                           "fragments": list(),
                           "verbose": False}

        # Initialize with default values
        self.__dict__ = default_options
        # Overwrite default values with user-provided ones, if they correspond to a valid keyword
        for k, v in opt_dict.items():
            if k in default_options:
                setattr(self, k, v)
            else:
                raise KeyError(f"Keyword :: {k}, not available in {self.__class__.__name__}.")

        # Raise error/warnings if input is not as expected
        if not self.geometry or not self.fragments:
            raise ValueError(f"A geometry and models must be provided when instantiating ONIOMProblemDecomposition.")

        self.geometry = atom_string_to_list(self.geometry) if isinstance(self.geometry, str) else self.geometry
        self.distribute_atoms()

        for frag in self.fragments:
            frag.build()

    def distribute_atoms(self):
        """For each fragment, the atom selection is passed to the Fragment
        object. Depending on the input, the method has several behaviors.
        """

        for fragment in self.fragments:
            # Case when no atom are selected -> whole system.
            if fragment.selected_atoms is None:
                fragment.geometry = self.geometry
            # Case where an int is detected -> first n atoms.
            elif type(fragment.selected_atoms) is int:
                fragment.geometry = self.geometry[:fragment.selected_atoms]
            # Case where a list of int is detected -> atom indexes are selected.
            # First atom is 0.
            elif isinstance(fragment.selected_atoms, list) and all(isinstance(id_atom, int) for id_atom in fragment.selected_atoms):
                fragment.geometry = [self.geometry[n] for n in fragment.selected_atoms]
            # Otherwise, an error is raised (list of float, str, etc.).
            else:
                raise TypeError("selected_atoms must be an int or a list of int.")

            # If there are broken_links (other than an empty list nor None).
            # The whole molecule geometry is needed to compute the position of
            # the capping atom (or functional group).
            if fragment.broken_links:
                for li in fragment.broken_links:
                    fragment.geometry += li.relink(self.geometry)

    def simulate(self):
        r"""Run the ONIOM core-method. The total energy is defined as
        E_ONIOM = E_LOW[SYSTEM] + \sum_i {E_HIGH_i[MODEL_i] - E_LOW_i[MODEL_i]}

        Returns:
            float: Total ONIOM energy.
        """

        # Run energy calculation for each fragment.
        e_oniom = sum([fragment.simulate() for fragment in self.fragments])

        return e_oniom

    def get_resources(self):
        """Estimate the resources required by ONIOM. Only supports fragments
        solved with quantum solvers. Resources for each fragments are outputed
        as a dictionary.
        """

        quantum_resources = dict()

        for fragment_i, fragment in enumerate(self.fragments):
            quantum_resources_fragment = fragment.get_resources()

            if quantum_resources_fragment:
                quantum_resources[fragment_i] = quantum_resources_fragment

                if self.verbose:
                    print(f"\t\tFragment Number : # {fragment_i + 1} \n\t\t{'-'*24}")
                    print(f"\t\t{quantum_resources[fragment_i]}\n")

        return quantum_resources



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from numpy import linspace

from tangelo.problem_decomposition.oniom.oniom_problem_decomposition import ONIOMProblemDecomposition
from tangelo.problem_decomposition.oniom._helpers.helper_classes import Fragment, Link
from tangelo.molecule_library import xyz_H4, xyz_PHE


class ONIOMTest(unittest.TestCase):

    def test_selected_atom(self):
        """Test selection of atoms. If it is not a list of int or an int, an
        error should be raised.
        """

        with self.assertRaises(TypeError):
            ONIOMProblemDecomposition({"geometry": xyz_H4,
                                       "fragments": [Fragment(solver_low="HF", solver_high="CCSD", selected_atoms=[3.1415])]})

    def test_not_implemented_solver(self):
        # Definition of simple fragments to test the error raising.

        with self.assertRaises(NotImplementedError):
            Fragment(solver_low="UNSUPPORTED")

        with self.assertRaises(NotImplementedError):
            Fragment(solver_low="HF",
                     solver_high="UNSUPPORTED",
                     selected_atoms=[0, 1])

    def test_capping_broken_link(self):
        """Testing the positon of a new H atom when a bond is broken."""

        system = Fragment(solver_low="HF")

        link = [Link(1, 2, 0.709, "H")]
        model = Fragment(solver_low="HF",
                         solver_high="CCSD",
                         selected_atoms=[0, 1, 9, 10, 11, 12, 13, 14, 22],
                         broken_links=link)

        oniom_solver = ONIOMProblemDecomposition({"geometry": xyz_PHE, "fragments": [system, model]})

        # Retrieving fragment geometry with an H atom replacing a broken bond.
        geom_fragment_capped = oniom_solver.fragments[1].geometry

        # Those position have been verified with a molecular visualization software.
        PHE_backbone_capped = [("N", (0.706, -1.9967, -0.0757)),
                               ("C", (1.1211, -0.6335, -0.4814)),
                               ("C", (2.6429, -0.5911, -0.5338)),
                               ("O", (3.1604, -0.2029, -1.7213)),
                               ("O", (3.4477, -0.8409, 0.3447)),
                               ("H", (-0.2916, -2.0354, -0.0544)),
                               ("H", (1.0653, -2.2124, 0.831)),
                               ("H", (0.699, -0.4698, -1.5067)),
                               ("H", (4.1118, -0.2131, -1.683)),
                               ("H", (0.772272, 0.1628488, 0.1778991))]

        # Every atom must be the same (same order too).
        # Position can be almost equals.
        for i, atom in enumerate(geom_fragment_capped):
            self.assertEqual(atom[0], PHE_backbone_capped[i][0])

            for dim in range(3):
                self.assertAlmostEqual(atom[1][dim], PHE_backbone_capped[i][1][dim], places=4)

    def test_energy_hf_ccsd_h4(self):
        """Testing the oniom energy with a low accuracy method (HF) and an
        higher one (CCSD) for H4 molecule. The H2-H2 interaction is computed at
        the HF level.
        """

        options_both = {"basis": "sto-3g"}

        system = Fragment(solver_low="HF", options_low=options_both)
        model_cc_1 = Fragment(solver_low="HF",
                              options_low=options_both,
                              solver_high="CCSD",
                              options_high=options_both,
                              selected_atoms=[0, 1])
        model_cc_2 = Fragment(solver_low="HF",
                              options_low=options_both,
                              solver_high="CCSD",
                              options_high=options_both,
                              selected_atoms=[2, 3])
        oniom_model_cc = ONIOMProblemDecomposition({"geometry": xyz_H4, "fragments": [system, model_cc_1, model_cc_2]})

        e_oniom_cc = oniom_model_cc.simulate()
        self.assertAlmostEqual(-1.901616, e_oniom_cc, places=5)

    def test_energy_hf_ccsd_phe(self):
        """Testing the oniom energy with a low accuracy method (HF) and an
        higher one (CCSD) for PHE molecule. The important fragment is chosen to
        be the backbone. The side chain is computed at the HF level.
        """

        options_low = {"basis": "sto-3g"}
        options_high = {"basis": "sto-3g"}

        system = Fragment(solver_low="HF", options_low=options_low)

        link = [Link(1, 2, 0.709, "H")]
        model = Fragment(solver_low="HF",
                         options_low=options_low,
                         solver_high="CCSD",
                         options_high=options_high,
                         selected_atoms=[0, 1, 9, 10, 11, 12, 13, 14, 22],
                         broken_links=link)

        oniom_solver = ONIOMProblemDecomposition({"geometry": xyz_PHE, "fragments": [system, model]})
        e_oniom = oniom_solver.simulate()

        self.assertAlmostEqual(e_oniom, -544.730619, places=4)

    def test_energy_multilayers(self):
        """Testing the oniom energy with a low accuracy method (HF), a medium
        accuracy (CCSD) and an higher one (FCI) for a H9 chain.
        """

        # H9 chain.
        xyz_h9 = [("H", (x, 0., 0.)) for x in linspace(-2., 2., num=9)]

        options = {"basis": "sto-3g"}

        # All system in HF.
        system = Fragment(solver_low="HF", options_low=options, spin=1)

        # Central 3 H energy is computed with FCI.
        high = Fragment(solver_low="CCSD",
                        options_low=options,
                        solver_high="FCI",
                        options_high=options,
                        selected_atoms=[3, 4, 5],
                        spin=1)

        # 2 H "buffer" atoms energy is computed with CCSD.
        medium = Fragment(solver_low="HF",
                          options_low=options,
                          solver_high="CCSD",
                          options_high=options,
                          selected_atoms=[2, 3, 4, 5, 6],
                          spin=1)

        oniom_solver = ONIOMProblemDecomposition({"geometry": xyz_h9, "fragments": [system, medium, high]})
        e_oniom = oniom_solver.simulate()

        self.assertAlmostEqual(-2.925695, e_oniom, places=5)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constructing guess orbitals in DMET calculation

Construction of the guess orbitals for fragment SCF calculation in DMET
calculation is done here.
"""

import scipy
import numpy as np
from functools import reduce


def dmet_fragment_guess(t_list, bath_orb, chemical_potential, norb_high, number_active_electron, active_fock):
    """Construct the guess orbitals.

    Args:
        t_list (list): Number of fragment & bath orbitals (int).
        bath_orb (numpy.array): The bath orbitals (float64).
        chemical_potential (float64): The chemical potential.
        norb_high (int): The number of orbitals in the fragment calculation.
        number_active_electron (int): The number of electrons in the fragment
            calculation.
        active_fock (numpy.array): The fock matrix from the low-level
            calculation (float64).

    Returns:
        numpy.array: The guess orbitals (float64).
    """

    # Construct the fock matrix of the fragment (subtract the chemical potential for consistency)
    fock_fragment = reduce(np.dot, (bath_orb[:, : norb_high].T, active_fock, bath_orb[:, : norb_high]))
    norb = t_list[0]
    if(chemical_potential != 0):
        for i in range(norb):
            fock_fragment[i, i] -= chemical_potential

    # Diagonalize the fock matrix and get the eigenvectors
    eigenvalues, eigenvectors = scipy.linalg.eigh(fock_fragment)
    eigenvectors = eigenvectors[:, eigenvalues.argsort()]

    # Extract the eigenvectors of the occupied orbitals as the guess orbitals
    frag_guess = np.dot(eigenvectors[ :, : int(number_active_electron/2)], eigenvectors[ :, : int(number_active_electron/2)].T) * 2

    return frag_guess



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class performing electronic structure calculation employing the
semi-empirical methods. At first, semi-empirical methods are ways of computing
the total energy of a molecule in a very fast way to optimize its geometry.
Those methods are not ab initio as they employ empirical parameters, as stated
in the name "semi-empirical". They are in fact related to simplified
Hartree-Fock versions with empirical corrections. Differences between them come
from the process chosen to compute the empirical parameters. For example, MINDO3
inventors used atomization energies to fit their mathematical models.

They have been introduced in this package for the purpose of computing an
environment energy and inducing constraints on atomic positions. As stand-alone
solvers, they are however a poor choice, as they do not provide an accurate
approximation of energies.

Here are the semi-empirical method(s) implemented:
    - MINDO3
"""

from tangelo.helpers.utils import is_package_installed
from tangelo.algorithms.electronic_structure_solver import ElectronicStructureSolver


class MINDO3Solver(ElectronicStructureSolver):
    """Uses the MINDO3 method to solve the electronic structure problem, through
    pyscf. Only the restricted (RMINDO3) flavor is implemented.

    Args:
        molecule (Molecule or SecondQuantizedMolecule): The molecule to
            simulate.

    Refs:
        - R. C. Bingham, M. J. Dewar, D. H. Lo, J. Am. Chem. Soc., 97, 1285
            (1975).
        - D. F. Lewis, Chem. Rev. 86, 1111 (1986).
    """

    def __init__(self, molecule):
        if not is_package_installed("pyscf.semiempirical"):
            raise ModuleNotFoundError(f"The pyscf.semiempirical module is not available and is required by {self.__class__.__name__}.")

        self.molecule = molecule

    def simulate(self):
        """Perform the simulation (energy calculation) for the molecule.

        Returns:
            float: RMINDO3 energy.
        """
        from pyscf.semiempirical import mindo3

        solver = mindo3.RMINDO3(self.molecule.to_pyscf()).run(verbose=0)
        total_energy = solver.e_tot

        return total_energy

    def get_rdm(self):
        """Method must be defined (ElectronicStructureSolver). For
        semi-empirical methods, it is not relevant nor defined.
        """

        raise NotImplementedError("Method get_rdm is not relevant for semi-empirical methods.")



import unittest

from scipy.linalg import expm
import numpy as np
from numpy.linalg import eigh
from openfermion import get_sparse_operator

from tangelo.linq import Simulator, Circuit, Gate
from tangelo.toolboxes.operators import FermionOperator, QubitOperator
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.molecule_library import mol_H4_sto3g
from tangelo.linq.tests.test_simulator import assert_freq_dict_almost_equal
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit
from tangelo.toolboxes.ansatz_generator.ansatz_utils import (givens_gate, trotterize, get_qft_circuit, controlled_swap_to_XX_gates,
                                                             derangement_circuit, controlled_pauliwords)

# Initiate simulators, Use cirq as it has the same ordering for statevectors as openfermion does for Hamiltonians
# This is important when converting the openfermion QubitOperator toarray(), propagating exactly and comparing
# to the statevector output of the simulator. All other simulators will produce the same statevector values but
# in a different order (i.e. msq_first instead of lsq_first)
sim = Simulator(target="cirq")

fermion_operator = mol_H4_sto3g._get_fermionic_hamiltonian()


class ansatz_utils_Test(unittest.TestCase):

    def test_trotterize_fermion_input(self):
        """ Verify that the time evolution is correct for different mappings and a fermionic
            hamiltonian input
        """

        time = 0.2
        for mapping in ["jw", "bk", "scbk"]:
            reference_circuit = get_reference_circuit(n_spinorbitals=mol_H4_sto3g.n_active_sos,
                                                      n_electrons=mol_H4_sto3g.n_active_electrons,
                                                      mapping=mapping,
                                                      up_then_down=True)
            _, refwave = sim.simulate(reference_circuit, return_statevector=True)

            qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=fermion_operator,
                                                         mapping=mapping,
                                                         n_spinorbitals=mol_H4_sto3g.n_active_sos,
                                                         n_electrons=mol_H4_sto3g.n_active_electrons,
                                                         up_then_down=True)

            ham_mat = get_sparse_operator(qubit_hamiltonian).toarray()
            evolve_exact = expm(-1j * time * ham_mat) @ refwave

            options = {"up_then_down": True,
                       "qubit_mapping": mapping,
                       "n_spinorbitals": mol_H4_sto3g.n_active_sos,
                       "n_electrons": mol_H4_sto3g.n_active_electrons}
            tcircuit, phase = trotterize(fermion_operator, trotter_order=1, n_trotter_steps=1, time=time,
                                         mapping_options=options, return_phase=True)
            _, wavefunc = sim.simulate(tcircuit, return_statevector=True, initial_statevector=refwave)
            wavefunc *= phase
            overlap = np.dot(np.conj(evolve_exact), wavefunc)
            self.assertAlmostEqual(overlap, 1.0, delta=1e-3)

    def test_trotterize_qubit_input(self):
        """ Verify that the time evolution is correct for different mappings and a qubit_hamiltonian input"""

        time = 0.2
        for mapping in ["jw", "bk", "scbk"]:
            reference_circuit = get_reference_circuit(n_spinorbitals=mol_H4_sto3g.n_active_sos,
                                                      n_electrons=mol_H4_sto3g.n_active_electrons,
                                                      mapping=mapping,
                                                      up_then_down=True)
            _, refwave = sim.simulate(reference_circuit, return_statevector=True)

            qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=fermion_operator,
                                                         mapping=mapping,
                                                         n_spinorbitals=mol_H4_sto3g.n_active_sos,
                                                         n_electrons=mol_H4_sto3g.n_active_electrons,
                                                         up_then_down=True)

            ham_mat = get_sparse_operator(qubit_hamiltonian).toarray()
            evolve_exact = expm(-1j * time * ham_mat) @ refwave

            tcircuit, phase = trotterize(qubit_hamiltonian, trotter_order=1, n_trotter_steps=1, time=time,
                                         return_phase=True)
            _, wavefunc = sim.simulate(tcircuit, return_statevector=True, initial_statevector=refwave)
            wavefunc *= phase
            overlap = np.dot(np.conj(evolve_exact), wavefunc)
            self.assertAlmostEqual(overlap, 1.0, delta=1e-3)

    def test_trotterize_different_order_and_steps(self):
        """ Verify that the time evolution is correct for different orders and number of steps
            with a qubit_hamiltonian input"""

        time = 0.2
        mapping = "bk"
        reference_circuit = get_reference_circuit(n_spinorbitals=mol_H4_sto3g.n_active_sos,
                                                  n_electrons=mol_H4_sto3g.n_active_electrons,
                                                  mapping=mapping,
                                                  up_then_down=True)
        _, refwave = sim.simulate(reference_circuit, return_statevector=True)

        qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=fermion_operator,
                                                     mapping=mapping,
                                                     n_spinorbitals=mol_H4_sto3g.n_active_sos,
                                                     n_electrons=mol_H4_sto3g.n_active_electrons,
                                                     up_then_down=True)

        ham_mat = get_sparse_operator(qubit_hamiltonian).toarray()
        evolve_exact = expm(-1j * time * ham_mat) @ refwave

        for trotter_order, n_trotter_steps in [(1, 1), (2, 1), (1, 2)]:

            tcircuit, phase = trotterize(qubit_hamiltonian, time, n_trotter_steps, trotter_order, return_phase=True)
            _, wavefunc = sim.simulate(tcircuit, return_statevector=True, initial_statevector=refwave)
            wavefunc *= phase
            overlap = np.dot(np.conj(evolve_exact), wavefunc)
            self.assertAlmostEqual(overlap, 1.0, delta=1e-3)
            # Test to make sure the same circuit is returned when return_phase=False.
            tcircuit = trotterize(qubit_hamiltonian, time, n_trotter_steps, trotter_order, return_phase=False)
            _, wavefunc = sim.simulate(tcircuit, return_statevector=True, initial_statevector=refwave)
            wavefunc *= phase
            overlap = np.dot(np.conj(evolve_exact), wavefunc)
            self.assertAlmostEqual(overlap, 1.0, delta=1e-3)

    def test_trotterize_fermionic_input_different_times(self):
        """ Verify that the time evolution is correct for a FermionOperator input with different times
            for each term
        """

        mapping = "jw"
        # generate Hermitian FermionOperator
        fermion_operators = [FermionOperator("0^ 3", 0.5) + FermionOperator("3^ 0", 0.5),
                             FermionOperator("1^ 2", 0.5) + FermionOperator("2^ 1", 0.5),
                             FermionOperator("1^ 3", 0.5) + FermionOperator("3^ 1", 0.5)]

        # time is twice as long as each Hermitian Operator has two terms
        time = {((0, 1), (3, 0)): 0.1, ((3, 1), (0, 0)): 0.1, ((1, 1), (2, 0)): 0.2, ((2, 1), (1, 0)): 0.2,
                ((1, 1), (3, 0)): 0.3, ((3, 1), (1, 0)): 0.3}

        # Build referenc circuit and obtain reference wavefunction
        reference_circuit = Circuit([Gate("X", 0), Gate("X", 3)])
        _, refwave = sim.simulate(reference_circuit, return_statevector=True)

        evolve_exact = refwave
        total_fermion_operator = FermionOperator()
        # evolve each term separately and apply to resulting wavefunction
        for i in range(3):
            total_fermion_operator += fermion_operators[i]
            qubit_hamiltonian = fermion_to_qubit_mapping(fermion_operator=fermion_operators[i],
                                                         mapping=mapping)
            ham_mat = get_sparse_operator(qubit_hamiltonian, n_qubits=4).toarray()
            evolve_exact = expm(-1j * time[next(iter(fermion_operators[i].terms))] * ham_mat) @ evolve_exact

        # Apply trotter-suzuki steps using different times for each term
        tcircuit, phase = trotterize(total_fermion_operator, trotter_order=1, n_trotter_steps=1, time=time, return_phase=True)
        _, wavefunc = sim.simulate(tcircuit, return_statevector=True, initial_statevector=refwave)
        wavefunc *= phase

        overlap = np.dot(np.conj(evolve_exact), wavefunc)
        self.assertAlmostEqual(overlap, 1.0, delta=1e-3)

    def test_trotterize_qubit_input_different_times(self):
        """ Verify that the time evolution is correct for a QubitOperator input with different times
            for each term
        """

        qubit_operator_list = [QubitOperator("X0 Y1", 0.5), QubitOperator("Y1 Z2", 0.5), QubitOperator("Y2 X3", 0.5)]

        time = {((0, 'X'), (1, 'Y')): 0.1, ((1, 'Y'), (2, 'Z')): 0.2, ((2, 'Y'), (3, 'X')): 0.3}

        # Generate initial wavefunction
        reference_circuit = Circuit([Gate("X", 0), Gate("X", 3)])
        _, refwave = sim.simulate(reference_circuit, return_statevector=True)

        # Exactly evolve for each time step
        evolve_exact = refwave
        for i in range(3):
            ham_mat = get_sparse_operator(qubit_operator_list[i], n_qubits=4).toarray()
            evolve_exact = expm(-1j * time[next(iter(qubit_operator_list[i].terms))] * ham_mat) @ evolve_exact

        # Apply trotter-suzuki with different times for each qubit operator term
        total_qubit_operator = QubitOperator()
        for qu_op in reversed(qubit_operator_list):
            total_qubit_operator += qu_op
        tcircuit, phase = trotterize(total_qubit_operator, trotter_order=2, n_trotter_steps=2, time=time, return_phase=True)
        _, wavefunc = sim.simulate(tcircuit, return_statevector=True, initial_statevector=refwave)
        wavefunc *= phase
        overlap = np.dot(np.conj(evolve_exact), wavefunc)
        self.assertAlmostEqual(overlap, 1.0, delta=1e-3)

    def test_qft_by_phase_estimation(self):
        """Test get_qft_circuit by applying phase-estimation to a 1-qubit operator with eigenvalue -1, i.e. phi=1/2"""
        n_qubits = 4
        qubit_list = [2, 1, 0]
        # Generate state with eigenvalue -1 of X operator exp(2*pi*i*phi) phi=1/2
        gate_list = [Gate("X", target=n_qubits-1), Gate("H", target=n_qubits-1)]

        # Generate phase-estimation circuit with three registers
        pe_circuit = Circuit(gate_list, n_qubits=n_qubits)
        qft = get_qft_circuit(qubit_list, n_qubits=n_qubits)
        pe_circuit += qft
        controlled_unitaries = []
        for i, qubit in enumerate(qubit_list):
            for j in range(2**i):
                controlled_unitaries += [Gate("CNOT", target=n_qubits-1, control=qubit)]
        pe_circuit += Circuit(controlled_unitaries, n_qubits=n_qubits)
        iqft = get_qft_circuit(qubit_list, n_qubits=n_qubits, inverse=True)
        pe_circuit += iqft

        # simulate starting state frequency is {"0001": 0.5, "0000": 0.5}
        freqs, _ = sim.simulate(pe_circuit)
        # phase is added to first three qubits with value  100 = 1 * 1/2 + 0 * 1/4 + 0 * 1/8
        # while keeping last qubit unchanged. Therefore, the target frequency dictionary is target_freq_dict
        target_freq_dict = {"1000": 0.5, "1001": 0.5}
        assert_freq_dict_almost_equal(target_freq_dict, freqs, atol=1.e-7)

    def test_controlled_time_evolution_by_phase_estimation(self):
        """ Verify that the time evolution is correct for a QubitOperator input with different times
            for each term
        """

        # Generate qubit operator with state 9 having eigenvalue 0.25
        qu_op = (QubitOperator("X0 X1", 0.125) + QubitOperator("Y1 Y2", 0.125) + QubitOperator("Z2 Z3", 0.125)
                 + QubitOperator("", 0.125))

        ham_mat = get_sparse_operator(qu_op).toarray()
        _, wavefunction = eigh(ham_mat)

        # Append four qubits in the zero state to eigenvector 9
        wave_9 = wavefunction[:, 9]
        for i in range(4):
            wave_9 = np.kron(wave_9, np.array([1, 0]))

        n_qubits = 8

        qubit_list = [7, 6, 5, 4]

        qft = get_qft_circuit(qubit_list, n_qubits=n_qubits)
        pe_circuit = qft
        for i, qubit in enumerate(qubit_list):
            u_circuit = trotterize(qu_op, trotter_order=1, n_trotter_steps=10, time=-2*np.pi, control=qubit)
            for j in range(2**i):
                pe_circuit += u_circuit
        iqft = get_qft_circuit(qubit_list, n_qubits=n_qubits, inverse=True)
        pe_circuit += iqft

        freqs, _ = sim.simulate(pe_circuit, initial_statevector=wave_9)

        # Trace out first 4 dictionary amplitudes, only care about final 4 indices
        trace_freq = dict()
        for key, value in freqs.items():
            trace_freq[key[-4:]] = trace_freq.get(key[-4:], 0) + value

        # State 9 has eigenvalue 0.25 so return should be 0100 (0*1/2 + 1*1/4 + 0*1/8 + 0*1/16)
        self.assertAlmostEqual(trace_freq["0100"], 1.0, delta=0.01)

    def test_controlled_swap(self):
        cswap_circuits = [Circuit([Gate("CSWAP", target=[1, 2], control=0)]),
                          Circuit(controlled_swap_to_XX_gates(0, 1, 2))]

        for cswap_circuit in cswap_circuits:
            # initialize in "110", returns "101"
            init_gates = [Gate("X", target=0), Gate("X", target=1)]
            circuit = Circuit(init_gates, n_qubits=3) + cswap_circuit
            freqs, _ = sim.simulate(circuit, return_statevector=True)
            assert_freq_dict_almost_equal({"101": 1.0}, freqs, atol=1.e-7)

            # initialize in "010" returns "010"
            init_gates = [Gate("X", target=1)]
            circuit = Circuit(init_gates, n_qubits=3) + cswap_circuit
            freqs, _ = sim.simulate(circuit, return_statevector=True)
            assert_freq_dict_almost_equal({"010": 1.0}, freqs, atol=1.e-7)

    def test_derangement_circuit_by_estimating_pauli_string(self):
        """ Verify that tr(rho^3 pa) for a pauliword pa is correct.
        Uses the exponential error suppression circuit
        """

        qu_op = QubitOperator("X0 Y1", 0.125) + QubitOperator("Y1 Y2", 0.125) + QubitOperator("Z2 Z3", 0.125)
        pa = QubitOperator("X0 X1 X2 X3", 1)

        ham_mat = get_sparse_operator(qu_op).toarray()
        _, wavefunction = eigh(ham_mat)
        pamat = get_sparse_operator(pa).toarray()

        mixed_wave = np.sqrt(3)/2*wavefunction[:, -1] + 1/2*wavefunction[:, 0]
        mixed_wave_3 = np.kron(np.kron(mixed_wave, mixed_wave), mixed_wave)
        full_start_vec = np.kron(mixed_wave_3, np.array([1, 0]))
        rho = np.outer(mixed_wave, mixed_wave)
        rho3 = rho @ rho @ rho
        exact = np.trace(rho3 @ pamat)

        n_qubits = 13

        qubit_list = [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11]]
        rho3_pa_circuit = Circuit([Gate("H", target=12)], n_qubits=n_qubits)
        derange_circuit = derangement_circuit(qubit_list, control=12, n_qubits=n_qubits)
        cpas = controlled_pauliwords(pa, control=12, n_qubits=n_qubits)
        rho3_pa_circuit += derange_circuit + cpas[0]
        rho3_pa_circuit += Circuit([Gate("H", target=12)], n_qubits=n_qubits)

        exp_op = QubitOperator("Z12", 1)
        measured = sim.get_expectation_value(exp_op, rho3_pa_circuit, initial_statevector=full_start_vec)
        self.assertAlmostEqual(measured, exact, places=6)

    def test_givens_gate(self):
        """Test of givens gate decomposition into 2 CNOTs and a CRY gate."""
        theta = 0.3

        # Explicit definition of givens rotation gate
        mat_rep = np.eye(4)
        mat_rep[1, 1] = np.cos(theta/2)
        mat_rep[1, 2] = -np.sin(theta/2)
        mat_rep[2, 1] = np.sin(theta/2)
        mat_rep[2, 2] = np.cos(theta/2)

        # Test that explicit definition and circuit return the same state vector
        vec = np.array([np.sqrt(2)/3, 2/3, np.sqrt(2)/3, 1/3])
        gvec = mat_rep@vec
        _, gvec2 = sim.simulate(Circuit(givens_gate([0, 1], theta)), return_statevector=True, initial_statevector=vec)
        np.testing.assert_array_almost_equal(gvec, gvec2)


if __name__ == "__main__":
    unittest.main()
